#cs Utility AutoIt3 Script
#
# EJECTUSB.AU3
#
# AI3 Script - EjectUSB - version 1.4
#
# Designed to be a simple utility to close all programs running from a specified drive or
# folder and then attempt ejection if a drive was specified. Extended functionality
# includes closing Explorer windows and removing registry entries and Recent Document
# shortcuts referencing the specified drive or folder.
#
# Prepared by William Hedrick
#
# Last compiled with AutoIt3 version 3.2.12.0 (ANSI)
# http://www.autoitscript.com/autoit3/
#
# Uses ProcessGetHWnd version 28th Dec 2006 by Helge
# http://www.autoitscript.com/fileman/users/Helge/udfs/_ProcessGetHWnd.au3
#
# Uses Hex2Dec version 16th Nov 2007 by SkinnyWhiteGuy
# http://www.autoitscript.com/forum/index.php?s=&showtopic=57258&view=findpost&p=433770
#
# Uses RefreshSystemTray version 1st Feb 2005 by Valik
# http://www.autoitscript.com/forum/index.php?showtopic=7404
#
# API-based ejection developed based on work by Uwe Sieber
# http://www.codeproject.com/KB/system/RemoveDriveByLetter.aspx
#
# WMI calls to detect drive removal based loosely on work by wraithdu
# http://www.autoitscript.com/forum/index.php?showtopic=83470
#
# [WH]
#
# No Copyright. No Company.
#ce=======================================================================================

;symbols

#NoTrayIcon

Const $text_header = "EjectUSB"
Const $text_detail = "The syntax for this utility is:"
Const $text_syntax = ".exe [DriveLetterOrFolder]"
Const $text_errors = "Invalid drive or folder specified."
Const $text_cancel = "Exit"
Const $text_ejectt = "Safe To Remove Hardware"
Const $text_ejectm = "The '%s' device can now be safely removed from the system."

;end

; ========================================================================================

;code

; Parse filename
$name = @ScriptName
If StringMid($name, StringLen($name) - 3, 1) = "." Then $name = StringTrimRight($name, 4)

; Parse program folder
$dir = @ScriptDir
If StringRight($dir, 1) <> "\" Then $dir &= "\"

; Initialize parameters
$cl_delay = 0
$cl_delay_value = ""
$cl_eject = 0
$cl_eject_value = 0
$cl_flush = 0
$cl_force = 0
$cl_nobat = 0
$cl_nomsg = 0
$file = ""

; Parse parameters
$i = 1
While $i <= $CmdLine[0]
	If $CmdLine[$i] = "/help" Or $CmdLine[$i] = "/h" Or $CmdLine[$i] = "/?" Or $CmdLine[$i] = "-h" Or $CmdLine[$i] = "-?" Then
		; Show help message
		MsgBox(64, $text_header, $text_detail & @CRLF & $name & $text_syntax)
		Exit 1
	ElseIf $CmdLine[$i] = "/delay" Then
		$cl_delay = 1
		If $i + 1 <= $CmdLine[0] And StringIsDigit($CmdLine[$i + 1]) = 1 Then
			$cl_delay_value = $CmdLine[$i + 1]
			$i += 1
		EndIf
	ElseIf $CmdLine[$i] = "/eject" Then
		$cl_eject = 1
		If $i + 1 <= $CmdLine[0] And StringIsDigit($CmdLine[$i + 1]) = 1 Then
			$cl_eject_value = Number($CmdLine[$i + 1])
			$i += 1
		EndIf
	ElseIf $CmdLine[$i] = "/flush" Then
		$cl_flush = 1
	ElseIf $CmdLine[$i] = "/force" Then
		$cl_force = 1
	ElseIf $CmdLine[$i] = "/nobat" Or $CmdLine[$i] = "/nobatch" Then
		$cl_nobat = 1
	ElseIf $CmdLine[$i] = "/nomsg" Then
		$cl_nomsg = 1
	ElseIf $file = "" And StringLeft($CmdLine[$i], 1) <> "/" And StringLeft($CmdLine[$i], 1) <> "-" Then
		$file = $CmdLine[$i]
	EndIf
	$i += 1
WEnd

; Prepare target location
If $file = "" Or $file = "\" Then
	; Find current drive letter if not specified
	If StringMid($dir, 2, 1) = ":" Then
		$file = StringLeft($dir, 2) & "\"
	Else
		$file = StringLeft($dir, StringInStr($dir, "\", 2, 3))
	EndIf
Else
	; Append :\ if single alphanumeric value
	If StringLen($file) = 1 And StringIsAlNum($file) = 1 Then $file &= ":\"

	; /force check
	If $cl_force = 0 And FileExists($file) = 0 Then
		; Abort if invalid value specified
		MsgBox(48, $text_header, $text_errors & @CRLF & @CRLF & $text_detail & @CRLF & $name & $text_syntax)
		Exit 1
	EndIf
EndIf

; Try to convert relative path into absolute path
If $file = ".." _
Or StringLeft($file, 3) = "..\" _
Or StringRight($file, 3) = "\.." _
Or StringInStr($file, "\..\", 2) > 0 _
Then $file = FileGetLongName($file, 1)

; Percorso di default in caso di errore (aggiunto Energy)
If @error Then 
	If StringMid($dir, 2, 1) = ":" Then
		$file = StringLeft($dir, 2) & "\"
	Else
		$file = StringLeft($dir, StringInStr($dir, "\", 2, 3))
	EndIf
EndIf

; Append backslash if missing
If StringRight($file, 1) <> "\" Then $file &= "\"

; ........................................................................................

; /flush check
If $cl_flush = 1 Then
	If IsAdmin() = 1 And StringLen($file) >= 3 And StringMid($file, 2, 1) = ":" Then
		; Prepare for DllCall()s
		$hDLLkernel32 = DllOpen("kernel32.dll")
		If $hDLLkernel32 <> -1 Then
			$hDLLOne = $hDLLkernel32
		Else
			$hDLLOne = "kernel32.dll"
		EndIf

		; Open volume handle, GENERIC_READ on a volume requires admin rights
		$flush_handle = DllCall($hDLLOne, "ptr", "CreateFile", "str", "\\.\" & StringLeft($file, 2), "dword", 0x80000000, "dword", 0x3, "ptr", 0, "dword", 3, "dword", 0, "ptr", 0)
		If @error <> 0 And $flush_handle[0] <> -1 Then
			; Flush file buffers
			DllCall($hDLLOne, "int", "FlushFileBuffers", "ptr", $flush_handle[0])
			If @error <> 0 Then
				; Close volume handle
				DllCall($hDLLOne, "int", "CloseHandle", "ptr", $flush_handle[0])

				; Close DLL handle
				If $hDLLkernel32 <> -1 Then DllClose($hDLLkernel32)

				; Quit successfully
				Exit
			EndIf

			; Close volume handle
			DllCall($hDLLOne, "int", "CloseHandle", "ptr", $flush_handle[0])
		EndIf

		; Close DLL handle
		If $hDLLkernel32 <> -1 Then DllClose($hDLLkernel32)
	EndIf

	; Quit with an errorlevel
	Exit 1
EndIf

; ........................................................................................

; /eject check
If $cl_eject = 1 Then
	If $cl_delay_value = "" Then
		$timetowait = "5"
	Else
		$timetowait = $cl_delay_value
	EndIf
	$donoteject = "0"
	If $cl_eject_value > 0 Then
		If ProcessWaitClose($cl_eject_value, Number($timetowait)) = 0 Then
			If ProcessExists($cl_eject_value) <> 0 Then
				ProcessClose($cl_eject_value)
				ProcessWaitClose($cl_eject_value, Number($timetowait))
			EndIf
		EndIf
	EndIf
	If $cl_delay = 1 Then
		$noejectapi = "1"
		Sleep(Number($timetowait) * 1000)
	Else
		$noejectapi = "0"
		If $cl_nomsg = 1 Then
			$nofakedmsg = "1"
		Else
			$nofakedmsg = "0"
		EndIf
	EndIf
Else

; ........................................................................................

; Read INI settings
If FileExists($dir & $name & ".ini") = 1 Then
	$timetowait = IniRead($dir & $name & ".ini", "Settings", "TimeToWait", "5")
	If StringIsDigit($timetowait) = 0 Then $timetowait = "5"
	$donotflush = IniRead($dir & $name & ".ini", "Settings", "DoNotFlush", "0")
	$donotclose = IniRead($dir & $name & ".ini", "Settings", "DoNotClose", "0")
	$noprograms = IniRead($dir & $name & ".ini", "Settings", "NoPrograms", "0")
	$nograceful = IniRead($dir & $name & ".ini", "Settings", "NoGraceful", "0")
	$noexplorer = IniRead($dir & $name & ".ini", "Settings", "NoExplorer", "0")
	$donotclean = IniRead($dir & $name & ".ini", "Settings", "DoNotClean", "0")
	$noregistry = IniRead($dir & $name & ".ini", "Settings", "NoRegistry", "0")
	$nofirewall = IniRead($dir & $name & ".ini", "Settings", "NoFirewall", "0")
	$noshortcut = IniRead($dir & $name & ".ini", "Settings", "NoShortcut", "0")
	$donotsubst = IniRead($dir & $name & ".ini", "Settings", "DoNotSubst", "0")
	$noredirect = IniRead($dir & $name & ".ini", "Settings", "NoRedirect", "0")
	$donoteject = IniRead($dir & $name & ".ini", "Settings", "DoNotEject", "0")
	$noejectapi = IniRead($dir & $name & ".ini", "Settings", "NoEjectAPI", "0")
	$nofakedmsg = IniRead($dir & $name & ".ini", "Settings", "NoFakedMsg", "0")
	$exclusions = IniReadSection($dir & $name & ".ini", "Exclusions")
	$excerror = @error
	$inclusions = IniReadSection($dir & $name & ".ini", "Inclusions")
	$incerror = @error

	If $excerror = 0 Then
		If $exclusions[0][0] = 0 Then $excerror = 1
	EndIf
	If $incerror = 0 Then
		If $inclusions[0][0] = 0 Then $incerror = 1
	EndIf
Else
	$timetowait = "5"
	$donotflush = "0"
	$donotclose = "0"
	$noprograms = "0"
	$nograceful = "0"
	$noexplorer = "0"
	$donotclean = "0"
	$noregistry = "0"
	$nofirewall = "0"
	$noshortcut = "0"
	$donotsubst = "0"
	$noredirect = "0"
	$donoteject = "0"
	$noejectapi = "0"
	$nofakedmsg = "0"
	$excerror = 1
	$incerror = 1
EndIf

$deadiconcheck = 0

If $nograceful = "0" Or $nograceful = "false" Then
	$gracefulcheck = 0
Else
	$gracefulcheck = 1
EndIf

; ........................................................................................

; /nobat check
If $cl_nobat = 0 Then

; BAT script
$batch_after = 0
If FileExists($dir & $name & ".bat") = 1 Then
	; Read first line
	$bat_line = FileReadLine($dir & $name & ".bat", 1)
	If @error = 0 Then
		; Save original target
		$bat_original = $file

		; Initialize input variables
		$batch_never = 0
		$batch_before = 0
		$batch_wait = 0
		Dim $batch_delregval[1] = [0]
		Dim $batch_delregkey[1] = [0]
		Dim $batch_delfile[1] = [0]
		$batch_quit = 0
		$batch_restore = 0

		; Remove trailing spaces
		If StringRight($bat_line, 1) = " " Then $bat_line = StringStripWS($bat_line, 2)

		; Check for valid keyword line
		If (StringLen($bat_line) > 5 And StringLeft($bat_line, 5) = "@rem ") _
		Or (StringLen($bat_line) > 3 And StringLeft($bat_line, 3) = ":: ") _
		Then
			; Format input
			$i = 1
			While $i <= StringLen($bat_line)
				If StringMid($bat_line, $i, 1) = '"' Then
					$i += 1
					While $i <= StringLen($bat_line)
						If StringMid($bat_line, $i, 1) = '"' Then ExitLoop
						$i += 1
					WEnd
				ElseIf StringMid($bat_line, $i, 1) = " " Then
					$bat_line = StringReplace($bat_line, $i, Chr(0))
				EndIf
				$i += 1
			WEnd
			$bat_line = StringReplace($bat_line, '"', "", 0, 2)
			$bat_line = StringReplace($bat_line, "%1", StringLeft($file, 1), 0, 2)
			$bat_line = StringReplace($bat_line, "%2", StringTrimLeft($file, 1), 0, 2)
			$bat_line = StringReplace($bat_line, "%3", StringLeft($dir, 1), 0, 2)
			$bat_line = StringReplace($bat_line, "%4", StringTrimLeft($dir, 1), 0, 2)
			$bat_list = StringSplit($bat_line, Chr(0))

			; Parse input
			$i = 2
			While $i <= $bat_list[0]
				If $bat_list[$i] = "nobat" Or $bat_list[$i] = "nobatch" Then
					$batch_never = 1
				ElseIf $bat_list[$i] = "after" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] = "resume" Then
						$batch_after = $i + 1
						$i += 1
					Else
						$batch_after = 2
					EndIf
				ElseIf $bat_list[$i] = "before" Then
					$batch_before = 1
				ElseIf $bat_list[$i] = "wait" Then
					$batch_wait = 1
				ElseIf $bat_list[$i] = "exit" Or $bat_list[$i] = "quit" Then
					$batch_quit = 1
				ElseIf $bat_list[$i] = "restore" Then
					$batch_restore = 1
				ElseIf $bat_list[$i] = "stop" Then
					ExitLoop
				ElseIf $bat_list[$i] = "jump" Then
					If $i + 1 <= $bat_list[0] And StringIsDigit($bat_list[$i + 1]) = 1 Then
						$i += Number($bat_list[$i + 1]) + 1
					EndIf
				ElseIf $bat_list[$i] = "sleep" Or $bat_list[$i] = "pause" Then
					If $i + 1 <= $bat_list[0] And StringIsDigit($bat_list[$i + 1]) = 1 Then
						Sleep(Number($bat_list[$i + 1]) * 1000)
						$i += 1
					Else
						Sleep(Number($timetowait) * 1000)
					EndIf
				ElseIf $bat_list[$i] = "eject" Then
					$cl_eject = 1
				ElseIf $bat_list[$i] = "force" Then
					$cl_force = 1
				ElseIf $bat_list[$i] = "noforce" Then
					$cl_force = 0
				ElseIf $bat_list[$i] = "target" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						If $bat_list[$i + 1] = "original" Then
							$file = $bat_original
						Else
							$file = $bat_list[$i + 1]
							If $file = "\" Then
								If StringMid($dir, 2, 1) = ":" Then
									$file = StringLeft($dir, 2) & "\"
								Else
									$file = StringLeft($dir, StringInStr($dir, "\", 2, 3))
								EndIf
							Else
								If StringLen($file) = 1 And StringIsAlNum($file) = 1 Then $file &= ":\"
								If $cl_force = 0 And FileExists($file) = 0 Then
									MsgBox(48, $text_header, $text_errors)
									Exit 1
								EndIf
							EndIf
							If $file = ".." Or StringLeft($file, 3) = "..\" Or StringRight($file, 3) = "\.." Or StringInStr($file, "\..\", 2) > 0 Then
								$file = FileGetLongName($file, 1)
							EndIf
							If StringRight($file, 1) <> "\" Then $file &= "\"
						EndIf
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "delregval" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" And StringInStr($bat_list[$i + 1], "\", 2) > 0 Then
						$batch_delregval[0] += 1
						ReDim $batch_delregval[$batch_delregval[0] + 1]
						$batch_delregval[$batch_delregval[0]] = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "delregkey" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" And StringInStr($bat_list[$i + 1], "\", 2) > 0 Then
						$batch_delregkey[0] += 1
						ReDim $batch_delregkey[$batch_delregkey[0] + 1]
						$batch_delregkey[$batch_delregkey[0]] = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "delfile" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$batch_delfile[0] += 1
						ReDim $batch_delfile[$batch_delfile[0] + 1]
						$batch_delfile[$batch_delfile[0]] = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "ifafter" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "ifdosname" Or $bat_list[$i] = "ifnotdosname" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						If $i + 2 <= $bat_list[0] And $bat_list[$i + 2] <> "" Then
							$bat_dos = "\Device\Error"
							If StringMid($file, 2, 1) = ":" And StringIsAlNum(StringLeft($file, 1)) = 1 Then
								$bat_ret = DllCall("kernel32.dll", "dword", "QueryDosDevice", "str", StringLeft($file, 2), "str", "", "dword", 260)
								If $bat_ret[0] <> 0 And UBound($bat_ret) >= 3 And $bat_ret[2] <> "" Then $bat_dos = $bat_ret[2]
								If @OSType = "WIN32_WINDOWS" And StringMid($bat_dos, 2, 1) = ":" Then
									If StringLen($bat_dos) > 2 Then
										$bat_dos = "\??\" & $bat_dos
									ElseIf StringLen($bat_dos = 2) Then
										$bat_type = DriveGetType($bat_dos & "\")
										If $bat_type = "Fixed" Then
											$bat_dos = "\Device\HarddiskVolume"
										ElseIf $bat_type = "CDROM" Then
											$bat_dos = "\Device\CdRom"
										ElseIf $bat_type = "Removable" Then
											$bat_dos = "\Device\Removable"
										ElseIf $bat_type = "Network" Then
											$bat_dos = "\Device\Network"
										EndIf
									EndIf
								EndIf
							ElseIf StringLeft($file, 2) = "\\" Then
								$bat_dos = "\Device\Shared"
							EndIf
							If StringInStr($bat_dos, $bat_list[$i + 1], 2) = 0 Then
								If $bat_list[$i] = "ifdosname" Then $i += 1
							Else
								If $bat_list[$i] = "ifnotdosname" Then $i += 1
							EndIf
						EndIf
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "ifexist" Or $bat_list[$i] = "ifnotexist" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						If $i + 2 <= $bat_list[0] And $bat_list[$i + 2] <> "" Then
							If FileExists($bat_list[$i + 1]) = 0 Then
								If $bat_list[$i] = "ifexist" Then $i += 1
							Else
								If $bat_list[$i] = "ifnotexist" Then $i += 1
							EndIf
						EndIf
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "ifwin9x" Or $bat_list[$i] = "ifnotwin9x" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						If @OSType = "WIN32_WINDOWS" Then
							If $bat_list[$i] = "ifnotwin9x" Then $i += 1
						Else
							If $bat_list[$i] = "ifwin9x" Then $i += 1
						EndIf
					EndIf
				ElseIf $bat_list[$i] = "ifosver" Or $bat_list[$i] = "ifnotosver" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						If $i + 2 <= $bat_list[0] And $bat_list[$i + 2] <> "" Then
							If @OSVersion = $bat_list[$i + 1] Then
								If $bat_list[$i] = "ifnotosver" Then $i += 1
							Else
								If $bat_list[$i] = "ifosver" Then $i += 1
							EndIf
						EndIf
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "ifcompname" Or $bat_list[$i] = "ifnotcompname" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						If $i + 2 <= $bat_list[0] And $bat_list[$i + 2] <> "" Then
							If @ComputerName = $bat_list[$i + 1] Then
								If $bat_list[$i] = "ifnotcompname" Then $i += 1
							Else
								If $bat_list[$i] = "ifcompname" Then $i += 1
							EndIf
						EndIf
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "ifusername" Or $bat_list[$i] = "ifnotusername" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						If $i + 2 <= $bat_list[0] And $bat_list[$i + 2] <> "" Then
							If @UserName = $bat_list[$i + 1] Then
								If $bat_list[$i] = "ifnotusername" Then $i += 1
							Else
								If $bat_list[$i] = "ifusername" Then $i += 1
							EndIf
						EndIf
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "timetowait" Then
					If $i + 1 <= $bat_list[0] And StringIsDigit($bat_list[$i + 1]) = 1 Then
						$timetowait = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "donotflush" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$donotflush = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "donotclose" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$donotclose = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "noprograms" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$noprograms = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "nograceful" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$nograceful = $bat_list[$i + 1]
						If $nograceful = "0" Or $nograceful = "false" Then
							$gracefulcheck = 0
						Else
							$gracefulcheck = 1
						EndIf
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "noexplorer" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$noexplorer = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "donotclean" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$donotclean = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "noregistry" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$noregistry = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "nofirewall" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$nofirewall = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "noshortcut" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$noshortcut = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "donotsubst" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$donotsubst = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "noredirect" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$noredirect = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "donoteject" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$donoteject = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "noejectapi" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$noejectapi = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "nofakedmsg" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$nofakedmsg = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "noexclusions" Then
					$excerror = 1
				ElseIf $bat_list[$i] = "noinclusions" Then
					$incerror = 1
				EndIf
				$i += 1
			WEnd
		EndIf

		; Run script
		If $batch_never = 0 And ($batch_after = 0 Or $batch_before = 1) Then
			If $batch_wait = 1 Then
				RunWait('"' & $dir & $name & '.bat" ' & StringLeft($file, 1) & ' "' & StringTrimLeft($file, 1) & '" ' & StringLeft($dir, 1) & ' "' & StringTrimLeft($dir, 1) & '"', "", @SW_HIDE)
			Else
				$bat_pid = Run('"' & $dir & $name & '.bat" ' & StringLeft($file, 1) & ' "' & StringTrimLeft($file, 1) & '" ' & StringLeft($dir, 1) & ' "' & StringTrimLeft($dir, 1) & '"', "", @SW_HIDE)
				If ProcessWaitClose($bat_pid, Number($timetowait)) = 0 Then
					If ProcessExists($bat_pid) <> 0 Then
						ProcessClose($bat_pid)
						ProcessWaitClose($bat_pid, Number($timetowait))
					EndIf
				EndIf
			EndIf
		EndIf

		; Delete specified registry values/keys and/or files
		If $batch_delregval[0] > 0 Or $batch_delregkey[0] > 0 Or $batch_delfile[0] > 0 Then
			$s = TimerInit()
			While $batch_delregval[0] > 0
				For $i = 1 To UBound($batch_delregval) - 1
					If $batch_delregval[$i] <> "" Then
						$bat_reg = StringInStr($batch_delregval[$i], "\", 2, -1)
						If $bat_reg = 0 Or RegDelete(StringLeft($batch_delregval[$i], $bat_reg - 1), StringTrimLeft($batch_delregval[$i], $bat_reg)) = 0 Then
							$batch_delregval[$i] = ""
							$batch_delregval[0] -= 1
						EndIf
					EndIf
				Next
				If $batch_delregval[0] <= 0 Or TimerDiff($s) > Number($timetowait) * 1000 Then ExitLoop
				Sleep(Opt("WinWaitDelay"))
			WEnd
			While $batch_delregkey[0] > 0
				For $i = 1 To UBound($batch_delregkey) - 1
					If $batch_delregkey[$i] <> "" Then
						If RegDelete($batch_delregkey[$i]) = 0 Then
							$batch_delregkey[$i] = ""
							$batch_delregkey[0] -= 1
						EndIf
					EndIf
				Next
				If $batch_delregkey[0] <= 0 Or TimerDiff($s) > Number($timetowait) * 1000 Then ExitLoop
				Sleep(Opt("WinWaitDelay"))
			WEnd
			While $batch_delfile[0] > 0
				For $i = 1 To UBound($batch_delfile) - 1
					If $batch_delfile[$i] <> "" Then
						If FileExists($batch_delfile[$i]) = 0 Then
							$batch_delfile[$i] = ""
							$batch_delfile[0] -= 1
						Else
							FileDelete($batch_delfile[$i])
						EndIf
					EndIf
				Next
				If $batch_delfile[0] <= 0 Or TimerDiff($s) > Number($timetowait) * 1000 Then ExitLoop
				Sleep(Opt("WinWaitDelay"))
			WEnd
		EndIf

		$batch_delregval = 0
		$batch_delregkey = 0
		$batch_delfile = 0

		If $batch_quit = 1 Then Exit
		If $batch_restore = 1 Then $file = $bat_original
	EndIf
EndIf

EndIf; /nobat check

; ........................................................................................

; Prevent system drive eject
If $file <> StringLeft(@WindowsDir, 3) And StringInStr($file, @WindowsDir, 2) = 0 Then

; ........................................................................................

; Flush Cached Data
If $donotflush = "0" Or $donotflush = "false" Then

If IsAdmin() = 1 Then

; Sync (Sysinternals)
If FileExists($dir & "sync.exe") = 1 And StringIsAlpha(StringLeft($file, 1)) = 1 Then
	$sync_version = FileGetVersion($dir & "sync.exe")
	; Check if registry keys already exist
	If $sync_version >= "2.2.0.0" Then
		RegRead("HKEY_CURRENT_USER\Software\Sysinternals", "")
		$sync_regone = @error
		RegRead("HKEY_CURRENT_USER\Software\Sysinternals\Sync", "")
		$sync_regtwo = @error
		RegWrite("HKEY_CURRENT_USER\Software\Sysinternals\Sync", "EulaAccepted", "REG_DWORD", 1)
	EndIf
	RunWait('"' & $dir & 'sync.exe" ' & StringLeft($file, 1), "", @SW_HIDE)
	; Remove registry keys if they didn't exist before
	If $sync_version >= "2.2.0.0" Then
		If $sync_regone <> 0 And $sync_regtwo <> 0 Then
			RegDelete("HKEY_CURRENT_USER\Software\Sysinternals")
		ElseIf $sync_regtwo <> 0 Then
			RegDelete("HKEY_CURRENT_USER\Software\Sysinternals\Sync")
		EndIf
	EndIf

; API Flush Code
ElseIf StringLen($file) >= 3 And StringMid($file, 2, 1) = ":" Then
	; Check if the situation needs a separate instance to avoid a stuck file handle
	If @OSType = "WIN32_NT" And ($donoteject = "0" Or $donoteject = "false") And StringLeft(@AutoItExe, 3) = StringLeft($file, 3) Then
		If StringRight(@ScriptName, 4) = ".exe" Then
			RunWait('"' & @ScriptFullPath & '" ' & StringLeft($file, 3) & ' /flush')
		Else
			RunWait('"' & @AutoItExe & '" "' & @ScriptFullPath & '" ' & StringLeft($file, 3) & ' /flush')
		EndIf
	Else
		; Prepare for DllCall()s
		$hDLLkernel32 = DllOpen("kernel32.dll")
		If $hDLLkernel32 <> -1 Then
			$hDLLOne = $hDLLkernel32
		Else
			$hDLLOne = "kernel32.dll"
		EndIf

		; Open volume handle, GENERIC_READ on a volume requires admin rights
		$flush_handle = DllCall($hDLLOne, "ptr", "CreateFile", "str", "\\.\" & StringLeft($file, 2), "dword", 0x80000000, "dword", 0x3, "ptr", 0, "dword", 3, "dword", 0, "ptr", 0)
		If @error <> 0 And $flush_handle[0] <> -1 Then
			; Flush file buffers
			DllCall($hDLLOne, "int", "FlushFileBuffers", "ptr", $flush_handle[0])
			; Close volume handle
			DllCall($hDLLOne, "int", "CloseHandle", "ptr", $flush_handle[0])
		EndIf

		; Close DLL handle
		If $hDLLkernel32 <> -1 Then DllClose($hDLLkernel32)
	EndIf
EndIf

EndIf

EndIf; End Flush Cached Data

; ........................................................................................

; Close Programs and Explorer
If $donotclose = "0" Or $donotclose = "false" Then

; ........................................................................................

; Close Programs
If $noprograms = "0" Or $noprograms = "false" Then

; OS Specific Closing
If @OSType = "WIN32_WINDOWS" Then

	; Close PStart portable Win98SE error window
	WinClose("PStart portable")

	; Prepare the data structure
	$tagPROCESSENTRY32 = DllStructCreate("dword;dword;dword;ulong_ptr;dword;dword;dword;long;dword;char[260]")
	DllStructSetData($tagPROCESSENTRY32, 1, DllStructGetSize($tagPROCESSENTRY32))

	; Prepare for DllCall()s
	$hDLLkernel32 = DllOpen("kernel32.dll")
	If $hDLLkernel32 <> -1 Then
		$hDLLOne = $hDLLkernel32
	Else
		$hDLLOne = "kernel32.dll"
	EndIf

; Graceful closing loop
While $gracefulcheck <= 1

; Process exclusions and inclusions
$aDLLCall = DllCall($hDLLOne, "ptr", "CreateToolhelp32Snapshot", "dword", 0x2, "dword", 0)
If @error = 0 Then
	$hProcessSnap = $aDLLCall[0]
	$aDLLCall = DllCall($hDLLOne, "int", "Process32First", "ptr", $hProcessSnap, "ptr", DllStructGetPtr($tagPROCESSENTRY32))
	If @error = 0 Then
		$r = Number($timetowait) * 1000
		While 1
			$loc = DllStructGetData($tagPROCESSENTRY32, 10)

	$k = 0
	If StringLeft($loc, StringLen($file)) = $file Then
		If DllStructGetData($tagPROCESSENTRY32, 3) <> @AutoItPID Then
			$k = 1
			; Process exclusions
			If $excerror = 0 Then
				For $j = 1 To $exclusions[0][0]
					If StringInStr($loc, $exclusions[$j][1], 2) > 0 Then $k = 0
				Next
			EndIf
		EndIf
	EndIf
	; Process inclusions
	If $incerror = 0 Then
		For $j = 1 To $inclusions[0][0]
			If StringInStr($loc, $inclusions[$j][1], 2) > 0 Then $k = 1
		Next
	EndIf
	; Terminate programs
	If $k = 1 Then
		If $nograceful = "0" Or $nograceful = "false" Then
			$prochwnd = _ProcessGetHWnd(DllStructGetData($tagPROCESSENTRY32, 3), 2, -1, 0)
			If $gracefulcheck = 1 Then
				If @error = 0 Then
					If $r <= 0 Then
						If ProcessExists(DllStructGetData($tagPROCESSENTRY32, 3)) <> 0 Then
							ProcessClose(DllStructGetData($tagPROCESSENTRY32, 3))
							$deadiconcheck = 1
						EndIf
					Else
						$s = TimerInit()
						If ProcessWaitClose(DllStructGetData($tagPROCESSENTRY32, 3), Ceiling($r / 1000)) = 0 Then
							If ProcessExists(DllStructGetData($tagPROCESSENTRY32, 3)) <> 0 Then
								ProcessClose(DllStructGetData($tagPROCESSENTRY32, 3))
								$deadiconcheck = 1
							EndIf
						EndIf
						$r -= TimerDiff($s)
					EndIf
				EndIf
			Else
				If @error = 0 Then
					For $j = 1 To $prochwnd[0][0]
						WinClose($prochwnd[$j][1])
					Next
					$gracefulcheck = -1
				Else
					If ProcessExists(DllStructGetData($tagPROCESSENTRY32, 3)) <> 0 Then
						ProcessClose(DllStructGetData($tagPROCESSENTRY32, 3))
						$deadiconcheck = 1
					EndIf
				EndIf
			EndIf
		Else
			If ProcessExists(DllStructGetData($tagPROCESSENTRY32, 3)) <> 0 Then
				ProcessClose(DllStructGetData($tagPROCESSENTRY32, 3))
				$deadiconcheck = 1
			EndIf
		EndIf
	EndIf

			$aDLLCall = DllCall("kernel32", "int", "Process32Next", "ptr", $hProcessSnap, "ptr", DllStructGetPtr($tagPROCESSENTRY32))
			If @error <> 0 Or $aDLLCall[0] = 0 Then ExitLoop
		WEnd
	EndIf
	DllCall($hDLLOne, "int", "CloseHandle", "ptr", $hProcessSnap)
EndIf

	$gracefulcheck += 2
WEnd

If $excerror = 0 Then

; Wait for programs to close on their own
$aDLLCall = DllCall($hDLLOne, "ptr", "CreateToolhelp32Snapshot", "dword", 0x2, "dword", 0)
If @error = 0 Then
	$hProcessSnap = $aDLLCall[0]
	$aDLLCall = DllCall($hDLLOne, "int", "Process32First", "ptr", $hProcessSnap, "ptr", DllStructGetPtr($tagPROCESSENTRY32))
	If @error = 0 Then
		$r = Number($timetowait) * 1000
		While 1
			$loc = DllStructGetData($tagPROCESSENTRY32, 10)

	$k = 0
	If StringLeft($loc, StringLen($file)) = $file Then
		If DllStructGetData($tagPROCESSENTRY32, 3) <> @AutoItPID Then
			$k = 1
			; Process exclusions
			For $j = 1 To $exclusions[0][0]
				If StringInStr($loc, $exclusions[$j][1], 2) > 0 Then $k = 0
			Next
		EndIf
	EndIf
	; Process inclusions
	If $incerror = 0 Then
		For $j = 1 To $inclusions[0][0]
			If StringInStr($loc, $inclusions[$j][1], 2) > 0 Then $k = 1
		Next
	EndIf
	; Wait for programs to close
	If $k = 1 Then
		If ProcessExists(DllStructGetData($tagPROCESSENTRY32, 3)) <> 0 Then
			$s = TimerInit()
			ProcessWaitClose(DllStructGetData($tagPROCESSENTRY32, 3), Ceiling($r / 1000))
			$r -= TimerDiff($s)
			If $r <= 0 Then ExitLoop
		EndIf
	EndIf

			$aDLLCall = DllCall("kernel32", "int", "Process32Next", "ptr", $hProcessSnap, "ptr", DllStructGetPtr($tagPROCESSENTRY32))
			If @error <> 0 Or $aDLLCall[0] = 0 Then ExitLoop
		WEnd
	EndIf
	DllCall($hDLLOne, "int", "CloseHandle", "ptr", $hProcessSnap)
EndIf

; Terminate programs
$aDLLCall = DllCall($hDLLOne, "ptr", "CreateToolhelp32Snapshot", "dword", 0x2, "dword", 0)
If @error = 0 Then
	$hProcessSnap = $aDLLCall[0]
	$aDLLCall = DllCall($hDLLOne, "int", "Process32First", "ptr", $hProcessSnap, "ptr", DllStructGetPtr($tagPROCESSENTRY32))
	If @error = 0 Then
		While 1
			$loc = DllStructGetData($tagPROCESSENTRY32, 10)

	$k = 0
	If StringLeft($loc, StringLen($file)) = $file Then
		If DllStructGetData($tagPROCESSENTRY32, 3) <> @AutoItPID Then
			$k = 1
		EndIf
	EndIf
	; Process inclusions
	If $incerror = 0 Then
		For $j = 1 To $inclusions[0][0]
			If StringInStr($loc, $inclusions[$j][1], 2) > 0 Then $k = 1
		Next
	EndIf
	; Terminate programs
	If $k = 1 Then
		If ProcessExists(DllStructGetData($tagPROCESSENTRY32, 3)) <> 0 Then
			ProcessClose(DllStructGetData($tagPROCESSENTRY32, 3))
			$deadiconcheck = 1
		EndIf
	EndIf

			$aDLLCall = DllCall("kernel32", "int", "Process32Next", "ptr", $hProcessSnap, "ptr", DllStructGetPtr($tagPROCESSENTRY32))
			If @error <> 0 Or $aDLLCall[0] = 0 Then ExitLoop
		WEnd
	EndIf
	DllCall($hDLLOne, "int", "CloseHandle", "ptr", $hProcessSnap)
EndIf

EndIf

; Wait for programs then close them
$aDLLCall = DllCall($hDLLOne, "ptr", "CreateToolhelp32Snapshot", "dword", 0x2, "dword", 0)
If @error = 0 Then
	$hProcessSnap = $aDLLCall[0]
	$aDLLCall = DllCall($hDLLOne, "int", "Process32First", "ptr", $hProcessSnap, "ptr", DllStructGetPtr($tagPROCESSENTRY32))
	If @error = 0 Then
		$r = Number($timetowait) * 1000
		While 1
			$loc = DllStructGetData($tagPROCESSENTRY32, 10)

	$k = 0
	If StringLeft($loc, StringLen($file)) = $file Then
		If DllStructGetData($tagPROCESSENTRY32, 3) <> @AutoItPID Then
			$k = 1
		EndIf
	EndIf
	; Process inclusions
	If $incerror = 0 Then
		For $j = 1 To $inclusions[0][0]
			If StringInStr($loc, $inclusions[$j][1], 2) > 0 Then $k = 1
		Next
	EndIf
	; Wait for programs to close
	If $k = 1 Then
		If $r <= 0 Then
			If ProcessExists(DllStructGetData($tagPROCESSENTRY32, 3)) <> 0 Then
				ProcessClose(DllStructGetData($tagPROCESSENTRY32, 3))
				$deadiconcheck = 1
			EndIf
		Else
			$s = TimerInit()
			If ProcessWaitClose(DllStructGetData($tagPROCESSENTRY32, 3), Ceiling($r / 1000)) = 0 Then
				If ProcessExists(DllStructGetData($tagPROCESSENTRY32, 3)) <> 0 Then
					ProcessClose(DllStructGetData($tagPROCESSENTRY32, 3))
					$deadiconcheck = 1
				EndIf
			EndIf
			$r -= TimerDiff($s)
		EndIf
	EndIf

			$aDLLCall = DllCall("kernel32", "int", "Process32Next", "ptr", $hProcessSnap, "ptr", DllStructGetPtr($tagPROCESSENTRY32))
			If @error <> 0 Or $aDLLCall[0] = 0 Then ExitLoop
		WEnd
	EndIf
	DllCall($hDLLOne, "int", "CloseHandle", "ptr", $hProcessSnap)
EndIf

	; Close DLL handle
	If $hDLLkernel32 <> -1 Then DllClose($hDLLkernel32)

	; Free the data structure
	$tagPROCESSENTRY32 = 0

Else; End Win9x Specific Section

	; Prepare for DllCall()s
	$hDLLkernel32 = DllOpen("kernel32.dll")
	If $hDLLkernel32 <> -1 Then
		$hDLLOne = $hDLLkernel32
	Else
		$hDLLOne = "kernel32.dll"
	EndIf
	$hDLLpsapi = DllOpen("psapi.dll")
	If $hDLLpsapi <> -1 Then
		$hDLLTwo = $hDLLpsapi
	Else
		$hDLLTwo = "psapi.dll"
	EndIf

; Graceful closing loop
While $gracefulcheck <= 1

; Process exclusions and inclusions
$r = Number($timetowait) * 1000
$list = ProcessList()
For $i = 1 To $list[0][0]
	$loc = _ProcessGetLocation($list[$i][1], $hDLLOne, $hDLLTwo)
	If @error <> 0 Then ContinueLoop
	$k = 0
	If StringLeft($loc, StringLen($file)) = $file Then
		If $list[$i][1] <> @AutoItPID Then
			$k = 1
			; Process exclusions
			If $excerror = 0 Then
				For $j = 1 To $exclusions[0][0]
					If StringInStr($loc, $exclusions[$j][1], 2) > 0 Then $k = 0
				Next
			EndIf
		EndIf
	EndIf
	; Process inclusions
	If $incerror = 0 Then
		For $j = 1 To $inclusions[0][0]
			If StringInStr($loc, $inclusions[$j][1], 2) > 0 Then $k = 1
		Next
	EndIf
	; Terminate programs
	If $k = 1 Then
		If $nograceful = "0" Or $nograceful = "false" Then
			$prochwnd = _ProcessGetHWnd($list[$i][1], 2, -1, 0)
			If $gracefulcheck = 1 Then
				If @error = 0 Then
					If $r <= 0 Then
						If ProcessExists($list[$i][1]) <> 0 Then
							ProcessClose($list[$i][1])
							$deadiconcheck = 1
						EndIf
					Else
						$s = TimerInit()
						If ProcessWaitClose($list[$i][1], Ceiling($r / 1000)) = 0 Then
							If ProcessExists($list[$i][1]) <> 0 Then
								ProcessClose($list[$i][1])
								$deadiconcheck = 1
							EndIf
						EndIf
						$r -= TimerDiff($s)
					EndIf
				EndIf
			Else
				If @error = 0 Then
					For $j = 1 To $prochwnd[0][0]
						WinClose($prochwnd[$j][1])
					Next
					$gracefulcheck = -1
				Else
					If ProcessExists($list[$i][1]) <> 0 Then
						ProcessClose($list[$i][1])
						$deadiconcheck = 1
					EndIf
				EndIf
			EndIf
		Else
			If ProcessExists($list[$i][1]) <> 0 Then
				ProcessClose($list[$i][1])
				$deadiconcheck = 1
			EndIf
		EndIf
	EndIf
Next

	$gracefulcheck += 2
WEnd

If $excerror = 0 Then

; Wait for programs to close on their own
$r = Number($timetowait) * 1000
$list = ProcessList()
For $i = 1 To $list[0][0]
	$loc = _ProcessGetLocation($list[$i][1], $hDLLOne, $hDLLTwo)
	If @error <> 0 Then ContinueLoop
	$k = 0
	If StringLeft($loc, StringLen($file)) = $file Then
		If $list[$i][1] <> @AutoItPID Then
			$k = 1
			; Process exclusions
			For $j = 1 To $exclusions[0][0]
				If StringInStr($loc, $exclusions[$j][1], 2) > 0 Then $k = 0
			Next
		EndIf
	EndIf
	; Process inclusions
	If $incerror = 0 Then
		For $j = 1 To $inclusions[0][0]
			If StringInStr($loc, $inclusions[$j][1], 2) > 0 Then $k = 1
		Next
	EndIf
	; Wait for programs to close
	If $k = 1 Then
		If ProcessExists($list[$i][1]) <> 0 Then
			$s = TimerInit()
			ProcessWaitClose($list[$i][1], Ceiling($r / 1000))
			$r -= TimerDiff($s)
			If $r <= 0 Then ExitLoop
		EndIf
	EndIf
Next

; Terminate programs
$list = ProcessList()
For $i = 1 To $list[0][0]
	$loc = _ProcessGetLocation($list[$i][1], $hDLLOne, $hDLLTwo)
	If @error <> 0 Then ContinueLoop
	$k = 0
	If StringLeft($loc, StringLen($file)) = $file Then
		If $list[$i][1] <> @AutoItPID Then
			$k = 1
			; Process exclusions (aggiunto da Energy)
			For $j = 1 To $exclusions[0][0]
				If StringInStr($loc, $exclusions[$j][1], 2) > 0 Then $k = 0
			Next
		EndIf
	EndIf
	; Process inclusions
	If $incerror = 0 Then
		For $j = 1 To $inclusions[0][0]
			If StringInStr($loc, $inclusions[$j][1], 2) > 0 Then $k = 1
		Next
	EndIf
	; Terminate programs
	If $k = 1 Then
		If ProcessExists($list[$i][1]) <> 0 Then
			ProcessClose($list[$i][1])
			$deadiconcheck = 1
		EndIf
	EndIf
Next

EndIf

; Wait for programs then close them
$r = Number($timetowait) * 1000
$list = ProcessList()
For $i = 1 To $list[0][0]
	$loc = _ProcessGetLocation($list[$i][1], $hDLLOne, $hDLLTwo)
	If @error <> 0 Then ContinueLoop
	$k = 0
	If StringLeft($loc, StringLen($file)) = $file Then
		If $list[$i][1] <> @AutoItPID Then
			$k = 1
			; Process exclusions (aggiunto da Energy)
			For $j = 1 To $exclusions[0][0]
				If StringInStr($loc, $exclusions[$j][1], 2) > 0 Then $k = 0
			Next
		EndIf
	EndIf
	; Process inclusions
	If $incerror = 0 Then
		For $j = 1 To $inclusions[0][0]
			If StringInStr($loc, $inclusions[$j][1], 2) > 0 Then $k = 1
		Next
	EndIf
	; Wait for programs to close
	If $k = 1 Then
		If $r <= 0 Then
			If ProcessExists($list[$i][1]) <> 0 Then
				ProcessClose($list[$i][1])
				$deadiconcheck = 1
			EndIf
		Else
			$s = TimerInit()
			If ProcessWaitClose($list[$i][1], Ceiling($r / 1000)) = 0 Then
				If ProcessExists($list[$i][1]) <> 0 Then
					ProcessClose($list[$i][1])
					$deadiconcheck = 1
				EndIf
			EndIf
			$r -= TimerDiff($s)
		EndIf
	EndIf
Next

; Attendi chiusura X-Launcher (aggiunto da Energy)
$list = ProcessList()
For $i = 1 To $list[0][0]
	$loc = _ProcessGetLocation($list[$i][1], $hDLLOne, $hDLLTwo)
	If @error <> 0 Then ContinueLoop
	$k = 0
	If StringLeft($loc, StringLen($file)) = $file Then
		If StringLeft($list[$i][0], 2) = 'X-' Then
			$k = 1
		EndIf
	EndIf
	; Wait for programs to close
	$r = Number($timetowait) * 1000
	If $k = 1 Then
		If ProcessExists($list[$i][1]) <> 0 Then
			If ProcessWaitClose($list[$i][1], Ceiling($r / 1000)) = 0 Then
				MsgBox(262192,"EjectUSB - Attention!","Manually close " & $list[$i][0] & " and press ok")
			EndIf
		EndIf
	EndIf
Next

	; Close DLL handles
	If $hDLLpsapi <> -1 Then DllClose($hDLLpsapi)
	If $hDLLkernel32 <> -1 Then DllClose($hDLLkernel32)

EndIf; End WinNT/XP Specific Section

EndIf; End Close Programs

; ........................................................................................

; Close explorer windows that are viewing the portable drive
If $noexplorer = "0" Or $noexplorer = "false" Then

$list = WinList("[CLASS:CabinetWClass]")
For $i = 1 To $list[0][0]
	If StringLeft(ControlGetText($list[$i][1], "", "[CLASSNN:Edit1]"), StringLen($file)) = $file Then WinClose($list[$i][1])
Next
$list = WinList("[CLASS:ExploreWClass]")
For $i = 1 To $list[0][0]
	If StringLeft(ControlGetText($list[$i][1], "", "[CLASSNN:Edit1]"), StringLen($file)) = $file Then WinClose($list[$i][1])
Next
$list = WinList("[CLASS:IEFrame]")
For $i = 1 To $list[0][0]
	If StringLeft(ControlGetText($list[$i][1], "", "[CLASSNN:Edit1]"), StringLen($file)) = $file Then WinClose($list[$i][1])
Next

EndIf; End Close Explorer

; ........................................................................................

; Close File Handles
If IsAdmin() = 1 Then

; Unlocker
If FileExists($dir & "Unlocker.exe") = 1 And FileExists($dir & "UnlockerDriver5.sys") = 1 _
And @OSType = "WIN32_NT" And @OSVersion <> "WIN_NT4" And @ProcessorArch = "X86" _
Then
	If StringInStr($file, " ", 2) > 0 Then
		$handle_file = FileGetShortName($file)
	Else
		$handle_file = $file
	EndIf
	RunWait('"' & $dir & 'Unlocker.exe" ' & $handle_file & ' /S')

; Handle (Sysinternals)
ElseIf FileExists($dir & "handle.exe") = 1 Then
	$handle_version = FileGetVersion($dir & "handle.exe")
	If ((@OSType = "WIN32_WINDOWS" Or @OSVersion = "WIN_NT4") And $handle_version <= "3.20.0.0") _
	Or (@OSType = "WIN32_NT" And @OSVersion <> "WIN_NT4") _
	Then
		$handle_exe = 'handle.exe" '
		; Check if registry keys already exist
		If $handle_version > "3.20.0.0" Then
			RegRead("HKEY_CURRENT_USER\Software\Sysinternals", "")
			$handle_regone = @error
			RegRead("HKEY_CURRENT_USER\Software\Sysinternals\Handle", "")
			$handle_regtwo = @error
			$handle_exe &= "/accepteula "
		EndIf
		If @OSType = "WIN32_WINDOWS" Then
			$handle_file = FileGetShortName($file)
		Else
			If StringInStr($file, " ", 2) > 0 Then
				$handle_file = '"' & $file & '"'
			Else
				$handle_file = $file
			EndIf
		EndIf
		$handle_output = ""
		$handle_pid = Run('"' & $dir & $handle_exe & $handle_file, "", @SW_HIDE, 2)
		; Capture output
		$s = TimerInit()
		While @error = 0
			$handle_output &= StdoutRead($handle_pid)
			If TimerDiff($s) > Number($timetowait) * 1000 Then ExitLoop
		WEnd
		; Remove registry keys if they didn't exist before
		If $handle_version > "3.20.0.0" Then
			If $handle_regone <> 0 And $handle_regtwo <> 0 Then
				RegDelete("HKEY_CURRENT_USER\Software\Sysinternals")
			ElseIf $handle_regtwo <> 0 Then
				RegDelete("HKEY_CURRENT_USER\Software\Sysinternals\Handle")
			EndIf
		EndIf
		; Parse output
		$handle_pid = ""
		$i = StringInStr($handle_output, @CRLF, 2)
		While $i > 0
			$i += 2
			$j = StringInStr($handle_output, @CRLF, 2, 1, $i)
			If $j = 0 Then ExitLoop
			; Grab current line
			$val = StringMid($handle_output, $i, $j - $i)
			$g = StringInStr($val, "pid: ", 2)
			If $g > 0 Then
				$g += 5
				$h = StringMid($val, $g, StringInStr($val, " ", 2, 1, $g) - $g)
				If @OSType = "WIN32_WINDOWS" Then $h = _HexToDec($h)
				If $h <> "" And $h <> $handle_pid Then
					$handle_exe = StringStripWS(StringLeft($val, $g - 6), 2)
					; Exclude self and system processes
					If $h = @AutoItPID Then
						$handle_pid = $h
					ElseIf @OSType = "WIN32_WINDOWS" And _
					(  $handle_exe = "kernel32.dll" _
					Or $handle_exe = "msgsrv32.exe" _
					Or $handle_exe = "mprexe.exe" _
					Or $handle_exe = "mmtask.tsk" _
					Or $handle_exe = "explorer.exe" _
					Or $handle_exe = "systray.exe" _
					Or $handle_exe = "ddhelp.exe" _
					Or $handle_exe = "wmiexe.exe" _
					)  Then
						$handle_pid = $h
					ElseIf @OSType = "WIN32_NT" And _
					(  $handle_exe = "System" _
					Or $handle_exe = "smss.exe" _
					Or $handle_exe = "csrss.exe" _
					Or $handle_exe = "winlogon.exe" _
					Or $handle_exe = "services.exe" _
					Or $handle_exe = "lsass.exe" _
					Or $handle_exe = "svchost.exe" _
					Or $handle_exe = "spoolsv.exe" _
					Or $handle_exe = "alg.exe" _
					Or $handle_exe = "explorer.exe" _
					)  Then
						$handle_pid = $h
					Else
						$loc = StringMid($val, StringInStr($val, ": ", 2, 1, $g) + 2)
						If @OSType = "WIN32_WINDOWS" Then $loc = FileGetLongName($loc)
						; Terminate programs
						If StringLeft($loc, StringLen($file)) = $file Then
							$prochwnd = _ProcessGetHWnd($h, 3, -1, 0)
							If @error = 0 Then
								For $k = 1 To $prochwnd[0][0]
									WinClose($prochwnd[$k][1])
								Next
							Else
								If ProcessExists($h) <> 0 Then
									ProcessClose($h)
									$deadiconcheck = 1
								EndIf
							EndIf
							$handle_pid = $h
						EndIf
					EndIf
				EndIf
			EndIf
			$i = $j
		WEnd
	EndIf
EndIf

EndIf; End Close File Handles

; ........................................................................................

; Remove dead system tray icons
If $deadiconcheck = 1 Then _RefreshSystemTray(Number($timetowait) * 1000)

EndIf; End Close Programs and Explorer

; ........................................................................................

; Stealth Cleanup
If $donotclean = "0" Or $donotclean = "false" Then

; Show Blinking Tray Icon
Opt("TrayMenuMode", 1)
TraySetIcon(@AutoItExe, 164)
TraySetToolTip(Chr(0))
TraySetState(5)
TraySetToolTip(Chr(0)); Fail-safe

; ........................................................................................

; Registry Cleanup
If $noregistry = "0" Or $noregistry = "false" Then

; MUICache (Vista)
$i = 1
While 1
	; Read value names
	$loc = RegEnumVal("HKEY_CURRENT_USER\Software\Classes\Local Settings\Software\Microsoft\Windows\Shell\MuiCache", $i)
	If @error <> 0 Then ExitLoop
	; Continue if no error
	$k = 0
	If StringLeft($loc, StringLen($file)) = $file Then $k = 1
	; Process inclusions
	If $incerror = 0 Then
		For $j = 1 To $inclusions[0][0]
			If StringInStr($loc, $inclusions[$j][1], 2) > 0 Then $k = 1
		Next
	EndIf
	; Process special exclusions
	If StringLeft($loc, 1) = "@" Or $loc = "LangID" Then $k = 0
	; Delete entries
	If $k = 1 Then
		If RegDelete("HKEY_CURRENT_USER\Software\Classes\Local Settings\Software\Microsoft\Windows\Shell\MuiCache", $loc) <> 1 Then $i += 1
	Else
		$i += 1
	EndIf
WEnd

; MUICache
$i = 1
While 1
	; Read value names
	$loc = RegEnumVal("HKEY_CURRENT_USER\Software\Microsoft\Windows\ShellNoRoam\MUICache", $i)
	If @error <> 0 Then ExitLoop
	; Continue if no error
	$k = 0
	If StringLeft($loc, StringLen($file)) = $file Then $k = 1
	; Process inclusions
	If $incerror = 0 Then
		For $j = 1 To $inclusions[0][0]
			If StringInStr($loc, $inclusions[$j][1], 2) > 0 Then $k = 1
		Next
	EndIf
	; Process special exclusions
	If StringLeft($loc, 1) = "@" Or $loc = "LangID" Then $k = 0
	; Delete entries
	If $k = 1 Then
		If RegDelete("HKEY_CURRENT_USER\Software\Microsoft\Windows\ShellNoRoam\MUICache", $loc) <> 1 Then $i += 1
	Else
		$i += 1
	EndIf
WEnd

; LastVisitedMRU
$i = 1
While 1
	; Read value names
	$loc = RegEnumVal("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\LastVisitedMRU", $i)
	If @error <> 0 Then ExitLoop
	; Continue if no error
	$val = RegRead("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\LastVisitedMRU", $loc)
	; Format string
	$val = _HexToString($val)
	$val = StringTrimLeft($val, StringInStr($val, Chr(0) & Chr(0), 2))
	$val = StringReplace($val, Chr(0), "", 0, 2)
	$k = 0
	If StringLeft($val, StringLen($file)) = $file Then $k = 1
	; Process inclusions
	If $incerror = 0 Then
		For $j = 1 To $inclusions[0][0]
			If StringInStr($val, $inclusions[$j][1], 2) > 0 Then $k = 1
		Next
	EndIf
	; Process special exclusions
	If $loc = "MRUList" Then $k = 0
	; Delete entries
	If $k = 1 Then
		If RegDelete("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\LastVisitedMRU", $loc) = 1 Then
			$val = RegRead("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\LastVisitedMRU", "MRUList")
			$val = StringReplace($val, $loc, "", 0, 2)
			RegWrite("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\LastVisitedMRU", "MRUList", "REG_SZ", $val)
		Else
			$i += 1
		EndIf
	Else
		$i += 1
	EndIf
WEnd

; LastVisitedPidlMRU
$i = 1
While 1
	; Read value names
	$loc = RegEnumVal("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\LastVisitedPidlMRU", $i)
	If @error <> 0 Then ExitLoop
	; Continue if no error
	$val = RegRead("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\LastVisitedPidlMRU", $loc)
	; Format string
	$val = _HexToString($val)
	$k = 0
	If StringInStr($val, $file, 2) > 0 Then $k = 1
	; Process inclusions
	If $incerror = 0 Then
		For $j = 1 To $inclusions[0][0]
			If StringInStr($val, $inclusions[$j][1], 2) > 0 Then $k = 1
		Next
	EndIf
	; Process special exclusions
	If $loc = "MRUListEx" Then $k = 0
	; Delete entries
	If $k = 1 And StringIsDigit($loc) = 1 Then
		If RegDelete("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\LastVisitedPidlMRU", $loc) = 1 Then
			$val = RegRead("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\LastVisitedPidlMRU", "MRUListEx")
			$val = StringReplace($val, Hex($loc, 2) & "000000", "", 0, 2)
			RegWrite("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\LastVisitedPidlMRU", "MRUListEx", "REG_BINARY", $val)
		Else
			$i += 1
		EndIf
	Else
		$i += 1
	EndIf
WEnd

; OpenSaveMRU
$h = 1
While 1
	; Read subkey names
	$key = RegEnumKey("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSaveMRU", $h)
	If @error <> 0 Then ExitLoop
	$i = 1
	While 1
		; Read value names
		$loc = RegEnumVal("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSaveMRU\" & $key, $i)
		If @error <> 0 Then ExitLoop
		; Continue if no error
		$val = RegRead("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSaveMRU\" & $key, $loc)
		$k = 0
		If StringLeft($val, StringLen($file)) = $file Then $k = 1
		; Process inclusions
		If $incerror = 0 Then
			For $j = 1 To $inclusions[0][0]
				If StringInStr($val, $inclusions[$j][1], 2) > 0 Then $k = 1
			Next
		EndIf
		; Process special exclusions
		If $loc = "MRUList" Then
			If $key <> "*" And $val = "" Then
				If RegDelete("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSaveMRU\" & $key) = 1 Then $i = 1
			EndIf
			$k = 0
		EndIf
		; Delete entries
		If $k = 1 Then
			If RegDelete("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSaveMRU\" & $key, $loc) = 1 Then
				$val = RegRead("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSaveMRU\" & $key, "MRUList")
				$val = StringReplace($val, $loc, "", 0, 2)
				If $key <> "*" And $val = "" Then
					If RegDelete("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSaveMRU\" & $key) = 1 Then $i = 1
				Else
					RegWrite("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSaveMRU\" & $key, "MRUList", "REG_SZ", $val)
				EndIf
			Else
				$i += 1
			EndIf
		Else
			$i += 1
		EndIf
	WEnd
	$h += 1
WEnd

; OpenSavePidlMRU
$h = 1
While 1
	; Read subkey names
	$key = RegEnumKey("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSavePidlMRU", $h)
	If @error <> 0 Then ExitLoop
	$i = 1
	While 1
		; Read value names
		$loc = RegEnumVal("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSavePidlMRU\" & $key, $i)
		If @error <> 0 Then ExitLoop
		; Continue if no error
		$val = RegRead("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSavePidlMRU\" & $key, $loc)
		; Format string
		$val = _HexToString($val)
		$k = 0
		If StringInStr($val, $file, 2) > 0 Then $k = 1
		; Process inclusions
		If $incerror = 0 Then
			For $j = 1 To $inclusions[0][0]
				If StringInStr($val, $inclusions[$j][1], 2) > 0 Then $k = 1
			Next
		EndIf
		; Process special exclusions
		If $loc = "MRUListEx" Then
			If $key <> "*" And $val = "" Then
				If RegDelete("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSavePidlMRU\" & $key) = 1 Then $i = 1
			EndIf
			$k = 0
		EndIf
		; Delete entries
		If $k = 1 And StringIsDigit($loc) = 1 Then
			If RegDelete("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSavePidlMRU\" & $key, $loc) = 1 Then
				$val = RegRead("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSavePidlMRU\" & $key, "MRUListEx")
				$val = StringReplace($val, Hex($loc, 2) & "000000", "", 0, 2)
				If $key <> "*" And $val = "ffffffff" Then
					If RegDelete("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSavePidlMRU\" & $key) = 1 Then $i = 1
				Else
					RegWrite("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSavePidlMRU\" & $key, "MRUListEx", "REG_BINARY", $val)
				EndIf
			Else
				$i += 1
			EndIf
		Else
			$i += 1
		EndIf
	WEnd
	$h += 1
WEnd

; RunMRU
$i = 1
While 1
	; Read value names
	$loc = RegEnumVal("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\RunMRU", $i)
	If @error <> 0 Then ExitLoop
	; Continue if no error
	$val = RegRead("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\RunMRU", $loc)
	$k = 0
	If StringLeft($val, StringLen($file)) = $file Then $k = 1
	; Process inclusions
	If $incerror = 0 Then
		For $j = 1 To $inclusions[0][0]
			If StringInStr($val, $inclusions[$j][1], 2) > 0 Then $k = 1
		Next
	EndIf
	; Process special exclusions
	If $loc = "MRUList" Then $k = 0
	; Delete entries
	If $k = 1 Then
		If RegDelete("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\RunMRU", $loc) = 1 Then
			$val = RegRead("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\RunMRU", "MRUList")
			$val = StringReplace($val, $loc, "", 0, 2)
			RegWrite("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\RunMRU", "MRUList", "REG_SZ", $val)
		Else
			$i += 1
		EndIf
	Else
		$i += 1
	EndIf
WEnd

; StreamMRU and Streams
$i = 1
While 1
	; Read value names
	$loc = RegEnumVal("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\StreamMRU", $i)
	If @error <> 0 Then ExitLoop
	; Continue if no error
	$val = RegRead("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\StreamMRU", $loc)
	; Format string
	$val = _HexToString($val)
	$k = 0
	If StringInStr($val, $file, 2) > 0 Then $k = 1
	; Process inclusions
	If $incerror = 0 Then
		For $j = 1 To $inclusions[0][0]
			If StringInStr($val, $inclusions[$j][1], 2) > 0 Then $k = 1
		Next
	EndIf
	; Process special exclusions
	If $loc = "MRUListEx" Then $k = 0
	; Delete entries
	If $k = 1 And StringIsDigit($loc) = 1 Then
		If RegDelete("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\StreamMRU", $loc) = 1 Then
			RegDelete("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\Streams\" & $loc)
			$val = RegRead("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\StreamMRU", "MRUListEx")
			$val = StringReplace($val, Hex($loc, 2) & "000000", "", 0, 2)
			RegWrite("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\StreamMRU", "MRUListEx", "REG_BINARY", $val)
		Else
			$i += 1
		EndIf
	Else
		$i += 1
	EndIf
WEnd

EndIf; End Registry Cleanup

; ........................................................................................

; Firewall Policy Cleanup
If $nofirewall = "0" Or $nofirewall = "false" Then

$i = 1
While 1
	; Read value names
	$loc = RegEnumVal("HKEY_LOCAL_MACHINE\System\CurrentControlSet\Services\SharedAccess\Parameters\FirewallPolicy\StandardProfile\AuthorizedApplications\List", $i)
	If @error <> 0 Then ExitLoop
	; Continue if no error
	$k = 0
	If StringLeft($loc, StringLen($file)) = $file Then $k = 1
	; Process inclusions
	If $incerror = 0 Then
		For $j = 1 To $inclusions[0][0]
			If StringInStr($loc, $inclusions[$j][1], 2) > 0 Then $k = 1
		Next
	EndIf
	; Delete entries
	If $k = 1 Then
		If RegDelete("HKEY_LOCAL_MACHINE\System\CurrentControlSet\Services\SharedAccess\Parameters\FirewallPolicy\StandardProfile\AuthorizedApplications\List", $loc) <> 1 Then
			$i += 1
		EndIf
	Else
		$i += 1
	EndIf
WEnd

EndIf; End Firewall Policy Cleanup

; ........................................................................................

; Recent Documents Cleanup
If $noshortcut = "0" Or $noshortcut = "false" Then

$val = RegRead("HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders", "Recent")
If @error = 0 Then
	If StringRight($val, 1) = "\" Then $val = StringTrimRight($val, 1)
	$list = _FastFileListToArray($val, "*.lnk", 1)
EndIf
If @error = 0 Then
	For $i = 1 To $list[0]
		; Read shortcut target
		$loc = FileGetShortcut($val & "\" & $list[$i])
		If @error = 0 Then
			$k = 0
			If StringLeft($loc[0], StringLen($file)) = $file _
			Or StringLeft($loc[1], StringLen($file)) = $file _
			Or StringLeft($loc[4], StringLen($file)) = $file _
			Then $k = 1
			; Process inclusions
			If $incerror = 0 Then
				For $j = 1 To $inclusions[0][0]
					If StringInStr($loc[0], $inclusions[$j][1], 2) > 0 _
					Or StringInStr($loc[1], $inclusions[$j][1], 2) > 0 _
					Or StringInStr($loc[4], $inclusions[$j][1], 2) > 0 _
					Then $k = 1
				Next
			EndIf
			If $k = 1 Then
				FileDelete($val & "\" & $list[$i])
			EndIf
		EndIf
	Next
EndIf

EndIf; End Recent Documents Cleanup

; ........................................................................................

; Hide Tray Icon
TraySetState(10)

EndIf; End Stealth Cleanup

; ........................................................................................

; Subst Support
If $donotsubst = "0" Or $donotsubst = "false" Then

; Prepare for DllCall()s
$hDLLkernel32 = DllOpen("kernel32.dll")
If $hDLLkernel32 <> -1 Then
	$hDLLOne = $hDLLkernel32
Else
	$hDLLOne = "kernel32.dll"
EndIf

; Check A to Z
For $i = 97 To 122
	$subst_dos = ""
	$subst_ret = DllCall($hDLLOne, "dword", "QueryDosDevice", "str", Chr($i) & ":", "str", "", "dword", 260)
	If $subst_ret[0] <> 0 And UBound($subst_ret) >= 3 And $subst_ret[2] <> "" Then $subst_dos = $subst_ret[2]
	If $subst_dos <> "" Then
		If StringLeft($subst_dos, 4) = "\??\" Then
			$subst_dos = StringTrimLeft($subst_dos, 4)
			If StringLeft($subst_dos, 4) = "UNC\" Then $subst_dos = "\" & StringTrimLeft($subst_dos, 3)
			If StringRight($subst_dos, 1) <> "\" Then $subst_dos &= "\"
		EndIf
		If StringLen($subst_dos) >= 3 And (StringMid($subst_dos, 2, 1) = ":" Or StringLeft($subst_dos, 2) = "\\") Then
			If StringRight($subst_dos, 1) <> "\" Then $subst_dos &= "\"
			; Un-Subst drive
			If StringLeft($subst_dos, StringLen($file)) = $file _
			Or Chr($i) & ":" = StringLeft($file, 2) _
			Then Run("subst.exe " & Chr($i) & ": /D", "", @SW_HIDE)
			; Redirect eject
			If $noredirect = "0" Or $noredirect = "false" Then
				If Chr($i) & ":" = StringLeft($file, 2) _
				And StringMid($subst_dos, 2, 1) = ":" _
				And StringLeft($subst_dos, 2) & "\" <> StringLeft(@WindowsDir, 3) _
				Then $file = StringLeft($subst_dos, 2) & "\"
			EndIf
		EndIf
	EndIf
Next

; Free array
$subst_ret = 0

; Close DLL handle
If $hDLLkernel32 <> -1 Then DllClose($hDLLkernel32)

EndIf; End Subst Support

; ........................................................................................

EndIf; Prevent system drive eject

; ........................................................................................

; /nobat check
If $cl_nobat = 0 Then

; BAT script (after)
If $batch_after > 0 Then
		; Reinitialize arrays
		Dim $batch_delregval[1] = [0]
		Dim $batch_delregkey[1] = [0]
		Dim $batch_delfile[1] = [0]

			; Parse input
			$i = $batch_after
			While $i <= $bat_list[0]
				If $bat_list[$i] = "nobat" Or $bat_list[$i] = "nobatch" Then
					$batch_never = 1
				ElseIf $bat_list[$i] = "wait" Then
					$batch_wait = 1
				ElseIf $bat_list[$i] = "nowait" Then
					$batch_wait = 0
				ElseIf $bat_list[$i] = "exit" Or $bat_list[$i] = "quit" Then
					$batch_quit = 1
				ElseIf $bat_list[$i] = "restore" Then
					$batch_restore = 1
				ElseIf $bat_list[$i] = "norestore" Then
					$batch_restore = 0
				ElseIf $bat_list[$i] = "stop" Then
					ExitLoop
				ElseIf $bat_list[$i] = "jump" Then
					If $i + 1 <= $bat_list[0] And StringIsDigit($bat_list[$i + 1]) = 1 Then
						$i += Number($bat_list[$i + 1]) + 1
					EndIf
				ElseIf $bat_list[$i] = "sleep" Or $bat_list[$i] = "pause" Then
					If $i + 1 <= $bat_list[0] And StringIsDigit($bat_list[$i + 1]) = 1 Then
						Sleep(Number($bat_list[$i + 1]) * 1000)
						$i += 1
					Else
						Sleep(Number($timetowait) * 1000)
					EndIf
				ElseIf $bat_list[$i] = "eject" Then
					$cl_eject = 1
				ElseIf $bat_list[$i] = "force" Then
					$cl_force = 1
				ElseIf $bat_list[$i] = "noforce" Then
					$cl_force = 0
				ElseIf $bat_list[$i] = "target" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						If $bat_list[$i + 1] = "original" Then
							$file = $bat_original
						Else
							$file = $bat_list[$i + 1]
							If $file = "\" Then
								If StringMid($dir, 2, 1) = ":" Then
									$file = StringLeft($dir, 2) & "\"
								Else
									$file = StringLeft($dir, StringInStr($dir, "\", 2, 3))
								EndIf
							Else
								If StringLen($file) = 1 And StringIsAlNum($file) = 1 Then $file &= ":\"
								If $cl_force = 0 And FileExists($file) = 0 Then
									MsgBox(48, $text_header, $text_errors)
									Exit 1
								EndIf
							EndIf
							If $file = ".." Or StringLeft($file, 3) = "..\" Or StringRight($file, 3) = "\.." Or StringInStr($file, "\..\", 2) > 0 Then
								$file = FileGetLongName($file, 1)
							EndIf
							If StringRight($file, 1) <> "\" Then $file &= "\"
						EndIf
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "delregval" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" And StringInStr($bat_list[$i + 1], "\", 2) > 0 Then
						$batch_delregval[0] += 1
						ReDim $batch_delregval[$batch_delregval[0] + 1]
						$batch_delregval[$batch_delregval[0]] = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "delregkey" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" And StringInStr($bat_list[$i + 1], "\", 2) > 0 Then
						$batch_delregkey[0] += 1
						ReDim $batch_delregkey[$batch_delregkey[0] + 1]
						$batch_delregkey[$batch_delregkey[0]] = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "delfile" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$batch_delfile[0] += 1
						ReDim $batch_delfile[$batch_delfile[0] + 1]
						$batch_delfile[$batch_delfile[0]] = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "ifbefore" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "ifdosname" Or $bat_list[$i] = "ifnotdosname" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						If $i + 2 <= $bat_list[0] And $bat_list[$i + 2] <> "" Then
							$bat_dos = "\Device\Error"
							If StringMid($file, 2, 1) = ":" And StringIsAlNum(StringLeft($file, 1)) = 1 Then
								$bat_ret = DllCall("kernel32.dll", "dword", "QueryDosDevice", "str", StringLeft($file, 2), "str", "", "dword", 260)
								If $bat_ret[0] <> 0 And UBound($bat_ret) >= 3 And $bat_ret[2] <> "" Then $bat_dos = $bat_ret[2]
								If @OSType = "WIN32_WINDOWS" And StringMid($bat_dos, 2, 1) = ":" Then
									If StringLen($bat_dos) > 2 Then
										$bat_dos = "\??\" & $bat_dos
									ElseIf StringLen($bat_dos = 2) Then
										$bat_type = DriveGetType($bat_dos & "\")
										If $bat_type = "Fixed" Then
											$bat_dos = "\Device\HarddiskVolume"
										ElseIf $bat_type = "CDROM" Then
											$bat_dos = "\Device\CdRom"
										ElseIf $bat_type = "Removable" Then
											$bat_dos = "\Device\Removable"
										ElseIf $bat_type = "Network" Then
											$bat_dos = "\Device\Network"
										EndIf
									EndIf
								EndIf
							ElseIf StringLeft($file, 2) = "\\" Then
								$bat_dos = "\Device\Shared"
							EndIf
							If StringInStr($bat_dos, $bat_list[$i + 1], 2) = 0 Then
								If $bat_list[$i] = "ifdosname" Then $i += 1
							Else
								If $bat_list[$i] = "ifnotdosname" Then $i += 1
							EndIf
						EndIf
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "ifexist" Or $bat_list[$i] = "ifnotexist" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						If $i + 2 <= $bat_list[0] And $bat_list[$i + 2] <> "" Then
							If FileExists($bat_list[$i + 1]) = 0 Then
								If $bat_list[$i] = "ifexist" Then $i += 1
							Else
								If $bat_list[$i] = "ifnotexist" Then $i += 1
							EndIf
						EndIf
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "ifwin9x" Or $bat_list[$i] = "ifnotwin9x" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						If @OSType = "WIN32_WINDOWS" Then
							If $bat_list[$i] = "ifnotwin9x" Then $i += 1
						Else
							If $bat_list[$i] = "ifwin9x" Then $i += 1
						EndIf
					EndIf
				ElseIf $bat_list[$i] = "ifosver" Or $bat_list[$i] = "ifnotosver" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						If $i + 2 <= $bat_list[0] And $bat_list[$i + 2] <> "" Then
							If @OSVersion = $bat_list[$i + 1] Then
								If $bat_list[$i] = "ifnotosver" Then $i += 1
							Else
								If $bat_list[$i] = "ifosver" Then $i += 1
							EndIf
						EndIf
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "ifcompname" Or $bat_list[$i] = "ifnotcompname" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						If $i + 2 <= $bat_list[0] And $bat_list[$i + 2] <> "" Then
							If @ComputerName = $bat_list[$i + 1] Then
								If $bat_list[$i] = "ifnotcompname" Then $i += 1
							Else
								If $bat_list[$i] = "ifcompname" Then $i += 1
							EndIf
						EndIf
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "ifusername" Or $bat_list[$i] = "ifnotusername" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						If $i + 2 <= $bat_list[0] And $bat_list[$i + 2] <> "" Then
							If @UserName = $bat_list[$i + 1] Then
								If $bat_list[$i] = "ifnotusername" Then $i += 1
							Else
								If $bat_list[$i] = "ifusername" Then $i += 1
							EndIf
						EndIf
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "timetowait" Then
					If $i + 1 <= $bat_list[0] And StringIsDigit($bat_list[$i + 1]) = 1 Then
						$timetowait = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "donoteject" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$donoteject = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "noejectapi" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$noejectapi = $bat_list[$i + 1]
						$i += 1
					EndIf
				ElseIf $bat_list[$i] = "nofakedmsg" Then
					If $i + 1 <= $bat_list[0] And $bat_list[$i + 1] <> "" Then
						$nofakedmsg = $bat_list[$i + 1]
						$i += 1
					EndIf
				EndIf
				$i += 1
			WEnd

		; Run script
		If $batch_never = 0 Then
			If $batch_wait = 1 Then
				RunWait('"' & $dir & $name & '.bat" ' & StringLeft($file, 1) & ' "' & StringTrimLeft($file, 1) & '" ' & StringLeft($dir, 1) & ' "' & StringTrimLeft($dir, 1) & '"', "", @SW_HIDE)
			Else
				$bat_pid = Run('"' & $dir & $name & '.bat" ' & StringLeft($file, 1) & ' "' & StringTrimLeft($file, 1) & '" ' & StringLeft($dir, 1) & ' "' & StringTrimLeft($dir, 1) & '"', "", @SW_HIDE)
				If ProcessWaitClose($bat_pid, Number($timetowait)) = 0 Then
					If ProcessExists($bat_pid) <> 0 Then
						ProcessClose($bat_pid)
						ProcessWaitClose($bat_pid, Number($timetowait))
					EndIf
				EndIf
			EndIf
		EndIf

		; Delete specified registry values/keys and/or files
		If $batch_delregval[0] > 0 Or $batch_delregkey[0] > 0 Or $batch_delfile[0] > 0 Then
			$s = TimerInit()
			While $batch_delregval[0] > 0
				For $i = 1 To UBound($batch_delregval) - 1
					If $batch_delregval[$i] <> "" Then
						$bat_reg = StringInStr($batch_delregval[$i], "\", 2, -1)
						If $bat_reg = 0 Or RegDelete(StringLeft($batch_delregval[$i], $bat_reg - 1), StringTrimLeft($batch_delregval[$i], $bat_reg)) = 0 Then
							$batch_delregval[$i] = ""
							$batch_delregval[0] -= 1
						EndIf
					EndIf
				Next
				If $batch_delregval[0] <= 0 Or TimerDiff($s) > Number($timetowait) * 1000 Then ExitLoop
				Sleep(Opt("WinWaitDelay"))
			WEnd
			While $batch_delregkey[0] > 0
				For $i = 1 To UBound($batch_delregkey) - 1
					If $batch_delregkey[$i] <> "" Then
						If RegDelete($batch_delregkey[$i]) = 0 Then
							$batch_delregkey[$i] = ""
							$batch_delregkey[0] -= 1
						EndIf
					EndIf
				Next
				If $batch_delregkey[0] <= 0 Or TimerDiff($s) > Number($timetowait) * 1000 Then ExitLoop
				Sleep(Opt("WinWaitDelay"))
			WEnd
			While $batch_delfile[0] > 0
				For $i = 1 To UBound($batch_delfile) - 1
					If $batch_delfile[$i] <> "" Then
						If FileExists($batch_delfile[$i]) = 0 Then
							$batch_delfile[$i] = ""
							$batch_delfile[0] -= 1
						Else
							FileDelete($batch_delfile[$i])
						EndIf
					EndIf
				Next
				If $batch_delfile[0] <= 0 Or TimerDiff($s) > Number($timetowait) * 1000 Then ExitLoop
				Sleep(Opt("WinWaitDelay"))
			WEnd
		EndIf

		$batch_delregval = 0
		$batch_delregkey = 0
		$batch_delfile = 0

		If $batch_quit = 1 Then Exit
		If $batch_restore = 1 Then $file = $bat_original
EndIf

EndIf; /nobat check

; ........................................................................................

EndIf; /eject check

; Drive Ejection
If $donoteject = "0" Or $donoteject = "false" Then

; Prevent system drive eject and validate target
If $file <> StringLeft(@WindowsDir, 3) And StringInStr($file, @WindowsDir, 2) = 0 _
And StringLen($file) = 3 And StringRight($file, 2) = ":\" _
Then
	If @OSType = "WIN32_NT" And $cl_eject = 0 And FileExists($dir & "EjectMedia.exe") = 1 Then
		Run('"' & $dir & 'EjectMedia.exe" ' & StringLeft($file, 2) & ' -L', "", @SW_MINIMIZE)
	ElseIf @OSType = "WIN32_NT" And $cl_eject = 0 And FileExists($dir & "RemoveDrive.exe") = 1 Then
		Run('"' & $dir & 'RemoveDrive.exe" ' & StringLeft($file, 2) & ' -L', "", @SW_MINIMIZE)
	ElseIf @OSType = "WIN32_NT" And $cl_eject = 0 And FileExists($dir & "USB_Disk_Eject.exe") = 1 Then
		Run('"' & $dir & 'USB_Disk_Eject.exe" /REMOVELETTER ' & StringLeft($file, 1))
	ElseIf @OSType = "WIN32_NT" And $cl_eject = 0 And FileExists($dir & "DevEject.exe") = 1 Then
		Run('"' & $dir & 'DevEject.exe" -EjectDrive:' & StringLeft($file, 2), "", @SW_HIDE)
	ElseIf @OSType = "WIN32_NT" And $cl_eject = 0 And FileExists($dir & "USBDeview.exe") = 1 Then
		Run('"' & $dir & 'USBDeview.exe" /stop_by_drive ' & StringLeft($file, 2))
	ElseIf @OSType = "WIN32_NT" And $cl_eject = 0 And FileExists($dir & "AutoStart.exe") = 1 Then
		Run('"' & $dir & 'AutoStart.exe" -no_main -eject ' & _GetDriveNumber(StringLeft($file, 1)))
	ElseIf @OSType = "WIN32_NT" And $cl_eject = 0 And FileExists($dir & "AutoEject.exe") = 1 Then
		Run('"' & $dir & 'AutoEject.exe" ' & _GetDriveNumber(StringLeft($file, 1)))
	Else
		If @OSType = "WIN32_NT" And StringRight(@ScriptName, 4) = ".exe" And StringLeft($dir, 3) = $file And StringLeft(@TempDir, 3) <> $file Then
			If StringInStr($name, " ", 2) > 0 Then $name = StringReplace($name, " ", "_", 0, 2)
			If StringInStr($name, '"', 2) > 0 Then $name = StringReplace($name, '"', '', 0, 2)
			If FileCopy(@ScriptFullPath, @TempDir & "\" & $name & ".exe", 9) = 1 Then
				; Prepare self deleting batch file
				$bat_file = @TempDir & "\" & $name & "_.bat"
				If FileExists($bat_file) = 0 _
				Or FileDelete($bat_file) = 1 _
				Then
					$bat_text = '@echo off' & @CRLF _
					& 'start /wait ' & $name & '.exe ' & $file _
					& ' /eject ' & @AutoItPID
					If Not ($noejectapi = "0" Or $noejectapi = "false") Then
						$bat_text &= ' /delay ' & $timetowait
					ElseIf Not ($nofakedmsg = "0" Or $nofakedmsg = "false") Then
						$bat_text &= ' /nomsg'
					EndIf
					$bat_text &= @CRLF _
					& 'attrib -R -S -H ' & $name & '.exe' & @CRLF _
					& 'del ' & $name & '.exe' & @CRLF _
					& 'attrib -R -S -H %0' & @CRLF _
					& '@cls' & @CRLF _
					& 'del %0'
					If FileWrite($bat_file, $bat_text) = 1 Then Run($bat_file, @TempDir, @SW_HIDE)
				EndIf
			EndIf
		Else
			If @OSType = "WIN32_NT" And ($noejectapi = "0" Or $noejectapi = "false") Then

; ........................................................................................

; Prepare Tray Icon
If $cl_eject = 0 And ($donotclean = "0" Or $donotclean = "false") Then
	TraySetIcon()
Else
	Opt("TrayMenuMode", 1)
	TraySetToolTip(Chr(0))
EndIf
; Prepare Tray Exit Option
Opt("TrayOnEventMode", 1)
Global $trayexit_value = 0
$trayexit = TrayCreateItem($text_cancel)
TrayItemSetOnEvent($trayexit, "_TrayExit")
; Show Tray Icon
TraySetState(1)
TraySetToolTip(Chr(0)); Fail-safe

; API Eject Code
$hDLLkernel32 = DllOpen("kernel32.dll")
If $hDLLkernel32 <> -1 Then
	$hDLLOne = $hDLLkernel32
Else
	$hDLLOne = "kernel32.dll"
EndIf

Do; Pseudo loop

; Make sure the drive letter exists
If $trayexit_value = 1 Or FileExists($file) = 0 Then ExitLoop

; Open volume handle
$ejectptr = DllCall($hDLLOne, "ptr", "CreateFile", "str", "\\.\" & StringLeft($file, 2), "dword", 0, "dword", 0x3, "ptr", 0, "dword", 3, "dword", 0, "ptr", 0)
; Check for function success
If @error <> 0 Or $ejectptr[0] = -1 Then ExitLoop

; Prepare the data structures
$ejectnum_struct = DllStructCreate("dword;ulong;ulong")
$ejectret_struct = DllStructCreate("dword")

; Get device number, IOCTL_STORAGE_GET_DEVICE_NUMBER = 0x002D1080
$ejectnum = -1
$ejectret = DllCall($hDLLOne, "int", "DeviceIoControl", "ptr", $ejectptr[0], "dword", 0x002D1080, "ptr", 0, "dword", 0, "ptr", DllStructGetPtr($ejectnum_struct), "dword", DllStructGetSize($ejectnum_struct), "dword*", DllStructGetPtr($ejectret_struct), "ptr", 0)
If @error = 0 And $ejectret[0] <> 0 Then $ejectnum = DllStructGetData($ejectnum_struct, 2)

; Close volume handle
DllCall($hDLLOne, "int", "CloseHandle", "ptr", $ejectptr[0])

; Make sure we got a device number
If $trayexit_value = 1 Or $ejectnum = -1 Then ExitLoop

; Get drive type
$ejectret = DllCall($hDLLOne, "uint", "GetDriveType", "str", $file)
$ejecttype = $ejectret[0]

; Get DOS device name
$ejectret = DllCall($hDLLOne, "dword", "QueryDosDevice", "str", StringLeft($file, 2), "str", "", "dword", 260)
; Check for function success
If $trayexit_value = 1 Or $ejectret[0] = 0 Or UBound($ejectret) < 3 Or $ejectret[2] = "" Then ExitLoop
$ejectdos = $ejectret[2]

; Get drive guid
If $ejecttype = 2 Then; DRIVE_REMOVABLE
	$ejectguid = DllStructCreate("long;short;short;byte[8]")
	If StringInStr($ejectdos, "\Floppy", 2) > 0 Then
		; GUID_DEVINTERFACE_FLOPPY = {53f56311-b6bf-11d0-94f2-00a0c91efb8b}
		DllStructSetData($ejectguid, 1, 0x53f56311)
	Else
		; GUID_DEVINTERFACE_DISK = {53f56307-b6bf-11d0-94f2-00a0c91efb8b}
		DllStructSetData($ejectguid, 1, 0x53f56307)
	EndIf
ElseIf $ejecttype = 3 Then; DRIVE_FIXED
	$ejectguid = DllStructCreate("long;short;short;byte[8]")
	; GUID_DEVINTERFACE_DISK = {53f56307-b6bf-11d0-94f2-00a0c91efb8b}
	DllStructSetData($ejectguid, 1, 0x53f56307)
ElseIf $ejecttype = 5 Then; DRIVE_CDROM
	$ejectguid = DllStructCreate("long;short;short;byte[8]")
	; GUID_DEVINTERFACE_CDROM = {53f56308-b6bf-11d0-94f2-00a0c91efb8b}
	DllStructSetData($ejectguid, 1, 0x53f56308)
Else
	ExitLoop
EndIf
DllStructSetData($ejectguid, 2, 0xb6bf)
DllStructSetData($ejectguid, 3, 0x11d0)
DllStructSetData($ejectguid, 4, 0x94, 1)
DllStructSetData($ejectguid, 4, 0xf2, 2)
DllStructSetData($ejectguid, 4, 0x00, 3)
DllStructSetData($ejectguid, 4, 0xa0, 4)
DllStructSetData($ejectguid, 4, 0xc9, 5)
DllStructSetData($ejectguid, 4, 0x1e, 6)
DllStructSetData($ejectguid, 4, 0xfb, 7)
DllStructSetData($ejectguid, 4, 0x8b, 8)

; Open DLL handle
$hDLLsetupapi = DllOpen("setupapi.dll")
If $hDLLsetupapi <> -1 Then
	$hDLLTwo = $hDLLsetupapi
Else
	$hDLLTwo = "setupapi.dll"
EndIf

; Open devinfo handle
$ejectinfo = DllCall($hDLLTwo, "ptr", "SetupDiGetClassDevs", "ptr", DllStructGetPtr($ejectguid), "ptr", 0, "hwnd", 0, "dword", 0x12)

; Check for valid devinfo handle
If $trayexit_value = 1 Or @error <> 0 Or $ejectinfo[0] = -1 Then
	If $hDLLsetupapi <> -1 Then DllClose($hDLLsetupapi)
	ExitLoop
EndIf

; Prepare data structures
$ejectdevinst = 0
$ejectbuf = DllStructCreate("byte[1024]")
$ejectbuf = DllStructGetSize($ejectbuf)
$ejectspdid = DllStructCreate("dword;long;short;short;byte[8];dword;dword")
DllStructSetData($ejectspdid, 1, DllStructGetSize($ejectspdid))
$ejectspdidd = DllStructCreate("dword;char[1024]")
DllStructSetData($ejectspdidd, 1, 5); dword + char = 5 bytes?

$i = -1
While 1
	$i += 1
	; Enumerate devices
	$ejectret = DllCall($hDLLTwo, "int", "SetupDiEnumDeviceInterfaces", "ptr", $ejectinfo[0], "ptr", 0, "ptr", DllStructGetPtr($ejectguid), "dword", $i, "ptr", DllStructGetPtr($ejectspdid))
	; Exit if no more devices
	If $trayexit_value = 1 Or @error <> 0 Or $ejectret[0] = 0 Then ExitLoop

	; Prepare a simple data structure
	$ejectsize = DllStructCreate("dword")

	; Get size value
	DllCall($hDLLTwo, "int", "SetupDiGetDeviceInterfaceDetail", "ptr", $ejectinfo[0], "ptr", DllStructGetPtr($ejectspdid), "ptr", 0, "dword", 0, "ptr", DllStructGetPtr($ejectsize), "ptr", 0)
	; Check for valid data
	If $trayexit_value = 1 Or @error <> 0 Then ExitLoop
	If DllStructGetData($ejectsize, 1) = 0 Or DllStructGetData($ejectsize, 1) > $ejectbuf Then ContinueLoop

	; Prepare data structure
	$ejectspdd = DllStructCreate("dword;long;short;short;byte[8];dword;dword")
	DllStructSetData($ejectspdd, 1, DllStructGetSize($ejectspdd))

	; Get interface details
	$ejectret = DllCall($hDLLTwo, "int", "SetupDiGetDeviceInterfaceDetail", "ptr", $ejectinfo[0], "ptr", DllStructGetPtr($ejectspdid), "ptr", DllStructGetPtr($ejectspdidd), "dword", DllStructGetData($ejectsize, 1), "ptr", DllStructGetPtr($ejectsize), "ptr", DllStructGetPtr($ejectspdd))
	; Check for function success
	If $trayexit_value = 1 Or @error <> 0 Then ExitLoop
	If $ejectret[0] = 0 Then ContinueLoop

	; Open drive handle
	$ejectptr = DllCall($hDLLOne, "ptr", "CreateFile", "str", DllStructGetData($ejectspdidd, 2), "dword", 0, "dword", 0x3, "ptr", 0, "dword", 3, "dword", 0, "ptr", 0)
	; Check for function success
	If $trayexit_value = 1 Or @error <> 0 Then ExitLoop
	If $ejectptr[0] = -1 Then ContinueLoop

	; Get device number, IOCTL_STORAGE_GET_DEVICE_NUMBER = 0x002D1080
	$ejectret = DllCall($hDLLOne, "int", "DeviceIoControl", "ptr", $ejectptr[0], "dword", 0x002D1080, "ptr", 0, "dword", 0, "ptr", DllStructGetPtr($ejectnum_struct), "dword", DllStructGetSize($ejectnum_struct), "dword*", DllStructGetPtr($ejectret_struct), "ptr", 0)
	; Check for function success and exit if a match
	If @error = 0 And $ejectret[0] <> 0 And DllStructGetData($ejectnum_struct, 2) = $ejectnum Then
		; Close drive handle
		DllCall($hDLLOne, "int", "CloseHandle", "ptr", $ejectptr[0])
		$ejectdevinst = DllStructGetData($ejectspdd, 6)
		ExitLoop
	EndIf

	; Close drive handle
	DllCall($hDLLOne, "int", "CloseHandle", "ptr", $ejectptr[0])
WEnd

; Close devinfo handle
DllCall($hDLLTwo, "int", "SetupDiDestroyDeviceInfoList", "ptr", $ejectinfo[0])

; Check for valid devinst
If $trayexit_value = 1 Or $ejectdevinst = 0 Then
	If $hDLLsetupapi <> -1 Then DllClose($hDLLsetupapi)
	ExitLoop
EndIf

; Get drive's parent
$ejectdevinstparent = DllStructCreate("dword")
DllCall($hDLLTwo, "int", "CM_Get_Parent", "ptr", DllStructGetPtr($ejectdevinstparent), "ptr", $ejectdevinst, "ulong", 0)
$ejectdevinstparent = DllStructGetData($ejectdevinstparent, 1)

; Check for valid devinstparent
If $trayexit_value = 1 Or $ejectdevinstparent = 0 Then
	If $hDLLsetupapi <> -1 Then DllClose($hDLLsetupapi)
	ExitLoop
EndIf

; Prepare data structure and empty string
$ejectvetotype = DllStructCreate("int")
$ejectvetoname = ""

; Loop until success, error or exit
While FileExists($file) = 1
	; Safely exit the loop before anything has happened
	If $trayexit_value = 1 Then ExitLoop

	; Try to eject the drive
	$ejectret = DllCall($hDLLTwo, "int", "CM_Request_Device_EjectW", "ptr", $ejectdevinstparent, "ptr", DllStructGetPtr($ejectvetotype), "wstr*", $ejectvetoname, "ulong", 260, "ulong", 0)

	; Exit on error or success
	If @error <> 0 Then
		ExitLoop
	ElseIf $ejectret[0] = 0 And DllStructGetData($ejectvetotype, 1) = 0 Then
		; Display a faked success message
		If $nofakedmsg = "0" Or $nofakedmsg = "false" Then
			; Prepare default success message
			$ejectmessage = $text_ejectm
			$ejecttitle = $text_ejectt

			; Get a handle to hotplug.dll (crashes if $hDLLOne is used in place of "kernel32.dll")
			$hDLLhotplug = DllCall("kernel32.dll", "hwnd", "LoadLibraryExA", "str", "hotplug.dll", "hwnd", 0, "int", 0x2)
			If @error = 0 Then
				; Open DLL handle
				$hDLLuser32 = DllOpen("user32.dll")
				If $hDLLuser32 <> -1 Then
					$hDLLTri = $hDLLuser32
				Else
					$hDLLTri = "user32.dll"
				EndIf

				; Load success message from hotplug.dll
				$ejectret = DllCall($hDLLTri, "int", "LoadString", "hwnd", $hDLLhotplug[0], "uint", 2260, "str", "", "int", 4096)
				If @error = 0 And IsArray($ejectret) And $ejectret[0] > 0 Then
					$ejectmessage = $ejectret[3]
				Else
					$ejectret = DllCall($hDLLTri, "int", "LoadString", "hwnd", $hDLLhotplug[0], "uint", 1260, "str", "", "int", 4096)
					If @error = 0 And IsArray($ejectret) And $ejectret[0] > 0 Then
						$ejectmessage = StringReplace($ejectret[3], "%s %s", "%s", 1, 2)
					EndIf
				EndIf
				$ejectret = DllCall($hDLLTri, "int", "LoadString", "hwnd", $hDLLhotplug[0], "uint", 2261, "str", "", "int", 4096)
				If @error = 0 And IsArray($ejectret) And $ejectret[0] > 0 Then
					$ejecttitle = $ejectret[3]
				Else
					$ejectret = DllCall($hDLLTri, "int", "LoadString", "hwnd", $hDLLhotplug[0], "uint", 1261, "str", "", "int", 4096)
					If @error = 0 And IsArray($ejectret) And $ejectret[0] > 0 Then
						$ejecttitle = $ejectret[3]
					EndIf
				EndIf

				; Close DLL handle
				If $hDLLuser32 <> -1 Then DllClose($hDLLuser32)
			EndIf

			; Replace %s with the drive letter
			$ejectmessage = StringReplace($ejectmessage, "%s", $file, 1, 2)
			; Replace \0 with nothing
			$ejectmessage = StringReplace($ejectmessage, Chr(0), "", 0, 2)
			; Check for an empty string
			If $ejectmessage = "" Then $ejectmessage = $file

			; Display the success message
			$ejectreg = RegRead("HKCU\Software\Microsoft\Windows\CurrentVersion\Explorer\Advanced", "EnableBalloonTips")
			If @OSType = "WIN32_NT" And @OSVersion <> "WIN_NT4" And (@error = -1 Or $ejectreg > 0) Then
				TrayTip($ejecttitle, $ejectmessage, 10, 1)
			Else
				Opt("WinSearchChildren", 1)
				$eject_hTray = WinGetHandle("[CLASS:TrayNotifyWnd]")
				If @error = 0 Then
					$eject_aTray = WinGetPos($eject_hTray)
					If @error = 0 Then
						If $eject_aTray[1] - 58 < 0 Then
							$eject_bTray = WinGetClientSize($eject_hTray)
							If @error = 0 Then
								$eject_aTray[1] = $eject_aTray[1] + $eject_bTray[1] + 6
								$eject_bTray = 0
							Else
								$eject_aTray[1] = $eject_aTray[1] + 26
							EndIf
						Else
							$eject_aTray[1] = $eject_aTray[1] - 58
						EndIf
					Else
						Dim $eject_aTray[2] = [@DesktopWidth, @DesktopHeight]
					EndIf
				Else
					Dim $eject_aTray[2] = [@DesktopWidth, @DesktopHeight]
				EndIf
				ToolTip($ejectmessage, $eject_aTray[0], $eject_aTray[1], $ejecttitle, 1, 4)
				$eject_aTray = 0
			EndIf

			; Detect Drive Removal
			$wmiSink = ObjCreate("WbemScripting.SWbemSink")
			If @error = 0 Then
				ObjEvent($wmiSink, "SINK_")
				If @error = 0 Then
					$Obj_WMIService = ObjGet('winmgmts:\\localhost\root\cimv2')
					If @error = 0 Then
						$obj_WMIService.ExecNotificationQueryAsync($wmiSink, "SELECT * FROM __InstanceOperationEvent WITHIN 1 WHERE TargetInstance ISA 'Win32_USBControllerdevice'")
					EndIf
				EndIf
			EndIf

			; Wait for the drive to be unplugged, an exit request or 10 seconds to pass
			$s = TimerInit()
			While 1
				If $trayexit_value = 1 Or TimerDiff($s) > 10000 Then ExitLoop 2
				Sleep(500)
			WEnd
		Else
			; Display the system's success message (typically causes an unnecessary delay)
			DllCall($hDLLTwo, "int", "CM_Request_Device_EjectW", "ptr", $ejectdevinstparent, "ptr", 0, "ptr", 0, "ulong", 0, "ulong", 0)
		EndIf
		ExitLoop
	EndIf

	; Safely exit the loop before sleeping
	If $trayexit_value = 1 Then ExitLoop
	Sleep(500)
WEnd

; Close DLL handle
If $hDLLsetupapi <> -1 Then DllClose($hDLLsetupapi)

Until 1

; Close DLL handle
If $hDLLkernel32 <> -1 Then DllClose($hDLLkernel32)

; Free all possible structures
$ejectnum_struct = 0
$ejectret_struct = 0
$ejectguid = 0
$ejectspdid = 0
$ejectspdidd = 0
$ejectsize = 0
$ejectspdd = 0
$ejectvetotype = 0

; Hide Tray Icon
If Opt("TrayIconHide") = 0 Then TraySetState(2)

; ........................................................................................

			Else

; ........................................................................................

; Eject Code
$ejectpid = Run("rundll32.exe shell32.dll,Control_RunDLL hotplug.dll", @SystemDir)
If @error = 0 Then $ejecthwnd = _ProcessGetHWnd($ejectpid, 3, -1, -1)
If @error = 0 Then
	For $g = 1 To $ejecthwnd[0][0]
		If ControlGetHandle($ejecthwnd[$g][1], "", 301) <> "" Then
			If WinWait($ejecthwnd[$g][1], "", Number($timetowait)) = 1 Then
				WinSetState($ejecthwnd[$g][1], "", @SW_HIDE)
				$checked = ControlCommand($ejecthwnd[$g][1], "", 307, "IsChecked", "")
				If $checked = 0 Then ControlCommand($ejecthwnd[$g][1], "", 307, "Check", "")
				$h = 0
				While ControlTreeView($ejecthwnd[$g][1], "", 301, "GetText", "#" & $h) <> ""
					$i = 0
					While ControlTreeView($ejecthwnd[$g][1], "", 301, "GetText", "#" & $h & "|#" & $i) <> ""
						If StringInStr(ControlTreeView($ejecthwnd[$g][1], "", 301, "GetText", "#" & $h & "|#" & $i), StringLeft($file, 2), 2) > 0 Then
							ControlTreeView($ejecthwnd[$g][1], "", 301, "Select", "#" & $h & "|#" & $i)
							If ControlClick($ejecthwnd[$g][1], "", 305) = 1 Then
								WinWaitNotActive($ejecthwnd[$g][1], "", Number($timetowait))
								$stophwnd = _ProcessGetHWnd($ejectpid, 3, -1, Number($timetowait) * 1000)
								If @error = 0 Then
									For $k = 1 To $stophwnd[0][0]
										If ControlGetHandle($stophwnd[$k][1], "", 321) <> "" Then
											If WinWait($stophwnd[$k][1], "", Number($timetowait)) = 1 Then
												ControlClick($stophwnd[$k][1], "", 1)
												WinClose($stophwnd[$k][1])
											EndIf
										EndIf
									Next
								EndIf
							EndIf
						EndIf
						$j = 0
						While ControlTreeView($ejecthwnd[$g][1], "", 301, "GetText", "#" & $h & "|#" & $i & "|#" & $j) <> ""
							If StringInStr(ControlTreeView($ejecthwnd[$g][1], "", 301, "GetText", "#" & $h & "|#" & $i & "|#" & $j), StringLeft($file, 2), 2) > 0 Then
								ControlTreeView($ejecthwnd[$g][1], "", 301, "Select", "#" & $h & "|#" & $i & "|#" & $j)
								If ControlClick($ejecthwnd[$g][1], "", 305) = 1 Then
									WinWaitNotActive($ejecthwnd[$g][1], "", Number($timetowait))
									$stophwnd = _ProcessGetHWnd($ejectpid, 3, -1, Number($timetowait) * 1000)
									If @error = 0 Then
										For $k = 1 To $stophwnd[0][0]
											If ControlGetHandle($stophwnd[$k][1], "", 321) <> "" Then
												If WinWait($stophwnd[$k][1], "", Number($timetowait)) = 1 Then
													ControlClick($stophwnd[$k][1], "", 1)
													WinClose($stophwnd[$k][1])
												EndIf
											EndIf
										Next
									EndIf
								EndIf
							EndIf
							$j += 1
						WEnd
						$i += 1
					WEnd
					$h += 1
				WEnd
				If $checked = 0 Then ControlCommand($ejecthwnd[$g][1], "", 307, "UnCheck", "")
				WinClose($ejecthwnd[$g][1])
			EndIf
		EndIf
	Next
EndIf

; ........................................................................................

			EndIf
		EndIf
	EndIf
EndIf

EndIf; End Drive Ejection

Exit

; ........................................................................................
; _ProcessGetLocation($iPID[, $hOne = "kernel32.dll"[, $hTwo = "psapi.dll"]])
;
; Returns the full path of an executable based on ProcessID (doesn't work on Win9x).
; Parameters:
;    $iPID - in/required - ProcessID
;    $hOne - in/optional - kernel32.dll handle
;    $hTwo - in/optional - psapi.dll handle
; Returns:
;    string containing full path of excutable
;    sets @error on failure and returns an empty string ""
; ........................................................................................
Func _ProcessGetLocation($iPID, $hOne = "kernel32.dll", $hTwo = "psapi.dll")
	Local $hProc = DllCall($hOne, "ptr", "OpenProcess", "dword", 0x0410, "int", 0, "dword", $iPID)
	If @error <> 0 Or $hProc[0] = 0 Then
		If $hProc[0] <> 0 Then DllCall($hOne, "int", "CloseHandle", "ptr", $hProc[0])
		Return SetError(1, 0, "")
	EndIf
	Local $stHMod = DllStructCreate("ptr")
	Local $stCB = DllStructCreate("dword")
	Local $resEnum = DllCall($hTwo, "int", "EnumProcessModules", "ptr", $hProc[0], "ptr", DllStructGetPtr($stHMod), "dword", DllStructGetSize($stHMod), "dword*", DllStructGetPtr($stCB, 1))
	If @error <> 0 Or $resEnum[0] = 0 Then
		DllCall($hOne, "int", "CloseHandle", "ptr", $hProc[0])
		Return SetError(2, 0, "")
	EndIf
	Local $resPath = DllCall($hTwo, "dword", "GetModuleFileNameEx", "ptr", $hProc[0], "ptr", DllStructGetData($stHMod, 1), "str", "", "dword", 2048)
	If @error <> 0 Or $resPath[0] = 0 Then
		DllCall($hOne, "int", "CloseHandle", "ptr", $hProc[0])
		Return SetError(3, 0, "")
	EndIf
	DllCall($hOne, "int", "CloseHandle", "ptr", $hProc[0])
	If UBound($resPath) < 4 Or StringLen($resPath[3]) = 0 Then Return SetError(4, 0, "")
	SetError(0)
	Return $resPath[3]
EndFunc; _ProcessGetLocation()

; ........................................................................................
; _ProcessGetHWnd($iPid[, $iOption = 1[, $sTitle = ""[, $iTimeout = 5000]]])
;
; Returns the HWND(s) owned by the specified process (PID only !).
; Parameter(s):
;    $iPid     - the owner-PID.
;    $iOption  - Optional : return/search methods :
;        0 - returns the HWND for the first non-titleless window.
;        1 - returns the HWND for the first found window (default).
;        2 - returns all HWNDs for all matches.
;        3 - returns all HWNDs for all visible matches.
;    $sTitle   - Optional : the title to match (see notes).
;    $iTimeout - Optional : timeout in msec (see notes)
; Return Value(s):
;    On Success - returns the HWND (see below for method 2).
;        $array[0][0] - number of HWNDs
;        $array[x][0] - title
;        $array[x][1] - HWND
;    On Failure - returns 0 and sets @error to 1.
; Note(s):
;    When a title is specified it will then only return the HWND to the titles matching
;    that specific string.
;    Use -1 to check all windows.
;
;    When using a timeout it's possible to use WinWaitDelay (Opt) to specify how often it
;    should wait before attempting another time to get the HWND.
;    Use -1 to loop until success or until PID doesn't exist.
;
; Author(s): Helge
;
; Modified to accept -1 for $sTitle and $iTimeout and added $iOption 3 [WH]
; ........................................................................................
Func _ProcessGetHWnd($iPid, $iOption = 1, $sTitle = "", $iTimeout = 5000)
	Local $aReturn[1][1] = [[0]], $aWin, $i, $hTimer = TimerInit()

	While ProcessExists($iPid) <> 0
		; Get list of windows
		If $sTitle = -1 Then
			$aWin = WinList()
		Else
			$aWin = WinList($sTitle)
		EndIf

		; Searches thru all windows
		For $i = 1 To $aWin[0][0]

			; Found a window owned by the given PID
			If $iPid = WinGetProcess($aWin[$i][1]) Then

				; Option 0 or 1 used
				If $iOption = 1 Or ($iOption = 0 And $aWin[$i][0] <> "") Then
					SetError(0)
					Return $aWin[$i][1]

				; Option 2 or 3 used
				ElseIf $iOption = 2 Or ($iOption = 3 And BitAnd(WinGetState($aWin[$i][1]), 2) <> 0) Then
					ReDim $aReturn[UBound($aReturn) + 1][2]
					$aReturn[0][0] += 1
					$aReturn[$aReturn[0][0]][0] = $aWin[$i][0]
					$aReturn[$aReturn[0][0]][1] = $aWin[$i][1]
				EndIf
			EndIf
		Next

		; If option 2 or 3 is used and there were matches then the list is returned
		If ($iOption = 2 Or $iOption = 3) And $aReturn[0][0] > 0 Then
			SetError(0)
			Return $aReturn
		EndIf

		; If timed out then give up
		If $iTimeout <> -1 And TimerDiff($hTimer) > $iTimeout Then ExitLoop

		; Waits before new attempt
		Sleep(Opt("WinWaitDelay"))
	WEnd

	; No matches
	SetError(1)
	Return 0
EndFunc; _ProcessGetHWnd()

; ........................................................................................
; Hex2Dec($hex)
;
; Converts a hex string to an unsigned integer.
; Parameters:
;    $hex - in/required - needs to be a hexadecimal string
; Returns:
;    integer
;
; Renamed to _HexToDec
; ........................................................................................
Func _HexToDec($hex)
	Local $dec = 0, $i
	For $i = 0 To StringLen($hex) - 1
		$dec += Dec(StringMid($hex, StringLen($hex)-$i,1)) * (16 ^ $i)
	Next
	Return $dec
EndFunc; _HexToDec()

; ........................................................................................
; _RefreshSystemTray([$nTimeout = 5000])
;
; Removes any dead icons from the system tray.
; Parameters:
;    $nTimeout - IN/OPTIONAL - The maximum time to wait for the notification area to
;    expand for Windows XP's "Hide Inactive Icons" feature (In milliseconds).
; Returns:
;    Sets @error on failure:
;        1 - Tray couldn't be found.
;        2 - DllCall error.
;
; Modified to use WM_MOUSEMOVE and changed $nDelay into $nTimeout [WH]
; ........................................................................................
Func _RefreshSystemTray($nTimeout = 5000)
	; Save WinSearchChildren setting
	Local $oldChildMode = Opt("WinSearchChildren", 1)

	Local $error = 0
	Do; Pseudo loop
		Local $hWnd = WinGetHandle("[CLASS:TrayNotifyWnd]")
		If @error <> 0 Then
			$error = 1
			ExitLoop
		EndIf

		; Account for differences between versions of Windows
		Local $hTray = ControlGetHandle($hWnd, "", "ToolbarWindow321")
		If $hTray = "" Then $hTray = $hWnd

		; Get initial system tray size
		Local $aTray = WinGetPos($hTray)
		If @error <> 0 Then
			$error = 1
			ExitLoop
		EndIf
		Local $bTray = $aTray

		; Check for the Hide Inactive Icons button
		Local $hHide = ControlGetHandle($hWnd, "", "Button1")
		If $hHide <> "" And ControlCommand($hWnd, "", $hHide, "IsVisible") Then 
			ControlClick($hWnd, "", $hHide)
			Local $hTimer = TimerInit()
			While 1
				Sleep(100)
				$bTray = WinGetPos($hTray)
				If @error <> 0 Or $bTray[2] = $aTray[2] Then
					ExitLoop
				ElseIf $bTray[2] < $aTray[2] Then
					ControlClick($hWnd, "", $hHide)
					$aTray[2] = $bTray[2]
					While 1
						Sleep(100)
						$bTray = WinGetPos($hTray)
						If @error <> 0 Or $bTray[2] = $aTray[2] Then ExitLoop 2
						$aTray[2] = $bTray[2]
						If TimerDiff($hTimer) > $nTimeout Then ExitLoop 2
					WEnd
				EndIf
				$aTray[2] = $bTray[2]
				If TimerDiff($hTimer) > $nTimeout Then ExitLoop
			WEnd

			; Get resulting system tray size
			$aTray = WinGetPos($hTray)
			If @error <> 0 Then
				$error = 1
				ExitLoop
			EndIf
		EndIf

		; Scrub the system tray
		Local $y = 0
		While $y < $aTray[3]
			Local $x = 0
			While $x < $aTray[2]
				; WM_MOUSEMOVE = 0x200
				DllCall("user32.dll", "none", "SendMessage", "hwnd", $hTray, "int", 0x200, "wparam", 0, "lparam", 65536 * $y + $x)
				If @error <> 0 Then
					$error = 2
					ExitLoop 2
				EndIf
				$bTray = WinGetPos($hTray)
				If @error = 0 And ($bTray[2] < $aTray[2] Or $bTray[3] < $aTray[3]) Then
					$x += 1
					$aTray = $bTray
				Else
					$x += 8
				EndIf
			WEnd
			$y += 8
		WEnd

		; Hide the inactive icons
		If $hHide <> "" And ControlCommand($hWnd, "", $hHide, "IsVisible") Then 
			ControlClick($hWnd, "", $hHide)
		EndIf
	Until 1

	; Restore WinSearchChildren setting
	Opt("WinSearchChildren", $oldChildMode)
	SetError($error)
EndFunc; _RefreshSystemTray()

; ........................................................................................
; Name...........: _HexToString
; Description ...: Convert a hex string to a string.
; Syntax.........: _HexToString($strHex)
; Parameters ....: $strHex - an hexadecimal string
; Return values .: Success - Returns a string.
;                  Failure - Returns -1 and sets @error to 1.
; Author ........: Jarvis Stubblefield
; Modified.......: 2005/09/04 jpm error checking
; Remarks .......: Copied out of String.au3
; ........................................................................................
Func _HexToString($strHex)
	Local $strChar, $aryHex, $i, $iDec, $Char, $iOne, $iTwo

	$aryHex = StringSplit($strHex, "")
	If Mod($aryHex[0], 2) <> 0 Then
		SetError(1)
		Return -1
	EndIf

	For $i = 1 To $aryHex[0]
		$iOne = $aryHex[$i]
		$i += 1
		$iTwo = $aryHex[$i]
		$iDec = Dec($iOne & $iTwo)
		If @error <> 0 Then
			SetError(1)
			Return -1
		EndIf
		
		$Char = Chr($iDec)
		$strChar &= $Char
	Next

	SetError(0)
	Return $strChar
EndFunc; _HexToString()

; ........................................................................................
; _FastFileListToArray($sPath[, $sFilter = "*"[, $iFlag = 0]])
;
; Based on RecursiveFileSearch (string based) by WeaponX (21st Feb 2008)
; http://www.autoitscript.com/forum/index.php?showtopic=58558
; Built to work in place of _FileListToArray() from File.au3
;
; Lists files and/or folders in a specified path (NOT recursively).
;
; Parameters:
;    $sPath   - in/required - Path to generate the file list for.
;    $sFilter - in/optional - The filter to use, default is *.
;    $iFlag   - in/optional - Specifies whether to return files, folders or both.
;    |$iFlag=0 (Default)    - Return both files and folders.
;    |$iFlag=1              - Return files only.
;    |$iFlag=2              - Return folders only.
; Returns:
;    The array returned is one-dimensional and is made up as follows:
;    $array[0] = Number of Files\Folders returned
;    $array[1] = 1st File\Folder
;    $array[2] = 2nd File\Folder
;    $array[3] = 3rd File\Folder
;    $array[n] = nth File\Folder
;    sets @error on failure and returns an empty string ""
;    |1 - Path not found or invalid
;    |2 - Invalid $sFilter
;    |3 - Invalid $iFlag
;    |4 - No File(s) Found
; ........................................................................................
Func _FastFileListToArray($sPath, $sFilter = "*", $iFlag = 0)
	Local $hSearch, $sFile, $sMaster = ""

	If Not FileExists($sPath) Then Return SetError(1, 1, "")
	If StringInStr($sFilter, "\") > 0 Or StringInStr($sFilter, "/") > 0 _
	Or StringInStr($sFilter, "<") > 0 Or StringInStr($sFilter, ">") > 0 _
	Or StringInStr($sFilter, ":") > 0 Or StringInStr($sFilter, "|") > 0 _
	Or StringStripWS($sFilter, 8) = "" Then Return SetError(2, 2, "")
	If Not ($iFlag = 0 Or $iFlag = 1 Or $iFlag = 2) Then Return SetError(3, 3, "")
	If StringRight($sPath, 1) = "\" Then $sPath = StringTrimRight($sPath, 1)

	$hSearch = FileFindFirstFile($sPath & "\" & $sFilter)
	If @error <> 0 Or $hSearch = -1 Then Return SetError(4, 4, "")

	While 1
		$sFile = FileFindNextFile($hSearch)
		If @error <> 0 Then ExitLoop
		If $iFlag = 1 And StringInStr(FileGetAttrib($sPath & "\" & $sFile), "D") > 0 Then ContinueLoop
		If $iFlag = 2 And StringInStr(FileGetAttrib($sPath & "\" & $sFile), "D") = 0 Then ContinueLoop
		$sMaster &= $sPath & "\" & $sFile & "*"
	WEnd
	FileClose($hSearch)
	If $sMaster = "" Then Return SetError(4, 4, "")

	Local $asFileList = StringSplit(StringTrimRight($sMaster, 1), "*")
	SetError(0)
	Return $asFileList
EndFunc; _FastFileListToArray()

; ........................................................................................
; _GetDriveNumber($iDrive)
;
; Converts drive letter to zero-based numeric index value.
; Parameters:
;    $iDrive - in/required - needs to be a single letter
; Returns:
;    string containing drive number
;    sets @error on failure and returns an empty string ""
; ........................................................................................
Func _GetDriveNumber($iDrive)
	If StringLen($iDrive) <> 1 Or StringIsAlpha($iDrive) <> 1 Then Return SetError(1, 0, "")
	SetError(0)
	If $iDrive = "A" Then Return "0"
	If $iDrive = "B" Then Return "1"
	If $iDrive = "C" Then Return "2"
	If $iDrive = "D" Then Return "3"
	If $iDrive = "E" Then Return "4"
	If $iDrive = "F" Then Return "5"
	If $iDrive = "G" Then Return "6"
	If $iDrive = "H" Then Return "7"
	If $iDrive = "I" Then Return "8"
	If $iDrive = "J" Then Return "9"
	If $iDrive = "K" Then Return "10"
	If $iDrive = "L" Then Return "11"
	If $iDrive = "M" Then Return "12"
	If $iDrive = "N" Then Return "13"
	If $iDrive = "O" Then Return "14"
	If $iDrive = "P" Then Return "15"
	If $iDrive = "Q" Then Return "16"
	If $iDrive = "R" Then Return "17"
	If $iDrive = "S" Then Return "18"
	If $iDrive = "T" Then Return "19"
	If $iDrive = "U" Then Return "20"
	If $iDrive = "V" Then Return "21"
	If $iDrive = "W" Then Return "22"
	If $iDrive = "X" Then Return "23"
	If $iDrive = "Y" Then Return "24"
	If $iDrive = "Z" Then Return "25"
	Return SetError(2, 0, "")
EndFunc; _GetDriveNumber()

; ........................................................................................
; _TrayExit()
;
; Tray Exit Option
; ........................................................................................
Func _TrayExit()
	Global $trayexit_value = 1
	; Hide Tray Icon
	TraySetState(2)
EndFunc; _TrayExit()

; ........................................................................................
; SINK_OnObjectReady()
;
; Detect Drive Removal
; ........................................................................................
Func SINK_OnObjectReady($objLatestEvent, $objAsyncContext)
    Switch $objLatestEvent.Path_.Class
        Case "__InstanceDeletionEvent"
            If StringInStr($objLatestEvent.TargetInstance.Dependent, 'DeviceID="USB\\', 2) > 0 Then
			Global $trayexit_value = 1
			ToolTip("")
			TrayTip("", "", 0)
			; Hide Tray Icon
			TraySetState(2)
            EndIf
    EndSwitch
EndFunc; SINK_OnObjectReady()

;end
